/*
 * Controller2d.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 * The base class for a 2d view's Controller.
 * This class is responsible for keeping the appropriate view focused
 */

abstract public class Controller2d 
        implements MouseMotionListener, 
                   MouseListener {
        
        protected Map _map;

        protected ViewControl _vc;
        
        protected Point3D _dragStartPt = new Point3D();
        protected boolean _rubberBanding = false;

        protected View2d _view = null;

        public Controller2d() {

        }
        
        /**
         * this modifies the Point p parameter
         * this handles the flip across the x axis and the translation
         * @param p is changed
         * @return the mutated p
         * 
         */
        abstract public Point3D screenToMap( Point3D p );
        
        public Point3D screenToMap( Point p ) {
                return screenToMap( new Point3D( p ) );
        }

        void setView( ViewControl vc, View2d view ) {
                _vc   = vc;
                _view = view;
        }

        void setMap( Map map ) {
                _map = map;
        }

        abstract protected class DragAction implements Map.ElementAction {
                Point3D diff;
                /** the Vector of verts to fill */
                Vector vertVect = new Vector();
        }
        
        /**
         * requests focus to the view and updates the ViewController info bar current info
         */
        public void mouseEntered( MouseEvent evt ) {
                _view.requestFocus();
                _vc.setInfoCurrent( _view, this );
        }

        /**
         * nulls the ViewController info bar current info
         */
        public void mouseExited( MouseEvent evt ) {
                _vc.setInfoCurrent( null, null );
        }

        abstract public void mouseMoved( MouseEvent evt );

        abstract public void mouseDragged( MouseEvent evt );
        
        abstract public void mousePressed( MouseEvent evt );
        
        abstract public void mouseReleased( MouseEvent evt );

        abstract public void mouseClicked( MouseEvent evt );

}

