/*
 * Controller2dSide.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

public class Controller2dSide extends Controller2d {

        View2dSide _viewSide = null;
        
        public Controller2dSide() {
                super();
        }
        
        void setView( ViewControl vc, View2d view ) {
                super.setView( vc, view );
                _viewSide = (View2dSide ) view;
        }

        /**
         * this modifies the Point p parameter
         * this handles the flip across the x axis and the translation
         * @param p is changed.  only x and y on the incoming p are meaningful
         * @return the mutated p
         * 
         */
        public Point3D screenToMap( Point3D p ) {
                p.x /= _view.getZoom();
                p.y /= _view.getZoom();

                Dimension d = _map.getMapSize();
                
                p.z = _map.getMapHeightZ()/2 - p.y;
                
                if ( _viewSide.getType() == _viewSide.HZ_X ) {
                        p.x = p.x - d.width / 2;
                        p.y = 0;
                }
                else {
                        p.y = p.x - d.height / 2;
                        p.x = 0;
                }
                
                return p;
        }

        public void mouseMoved( MouseEvent evt ) {
                
                Point scr_p = evt.getPoint();
                
                Point3D p = screenToMap( scr_p );

                _vc.setInfoPointer( p );
        }
        
        private boolean drag_floor_ok = true;
        private boolean drag_ceil_ok = true;
        
        DragAction _dragAction = new DragAction() {
                        public void doAction( Map.Element elem ) {
                                
                                if ( elem instanceof Map.LightSource ) {
                                        _map.moveLightRelative( (Map.LightSource) elem, diff, true );
                                        return;
                                }
                                
                                if ( elem instanceof Map.Sector ) {

                                        Map.Sector s = (Map.Sector) elem;
                                        Map.Vertex[] svs = s.s_vertexes;
                                        
                                        for ( int i = 0; i < svs.length; ++i ) {
                                                if ( ! vertVect.contains( svs[i] ) )
                                                        vertVect.addElement( svs[i] ); 
                                        }
                                        
                                        if ( drag_floor_ok )
                                                s.floorHeight += diff.z;

                                        if ( drag_ceil_ok )
                                                s.ceilingHeight += diff.z;
                                        
                                        return;
                                }
                        }
                };
        
        public void mouseDragged( MouseEvent evt ) {
                
                Point3D p = screenToMap( evt.getPoint() );
                
                _vc.setInfoPointer( p );
                
                int mod = evt.getModifiers();
                
                //
                // middle button down, drag the items
                //

                if ( ( mod & evt.BUTTON2_MASK ) != 0 && _vc.isMode( _vc.MODE_SELECT ) && _map.hasSelections() ) {

                        Point3D diff = new Point3D();
                        
                        int gridSize = _vc.getGridSize();
                        int halfGrid = gridSize / 2;
                        
                        Map.Element elem = _map.getLastSelection();
                        
                        if ( elem == null )
                                return;

                        diff.x = p.x - _dragStartPt.x;
                        diff.y = p.y - _dragStartPt.y;
                        diff.z = p.z - _dragStartPt.z;
                        
                        //
                        // snap to closest point if needed
                        //

                        if ( _vc.isSnapEnabled() ) {
                                if (diff.x >= halfGrid ) {
                                        diff.x = ( ( diff.x + halfGrid ) / gridSize ) * gridSize;
                                }
                                else if (diff.x <= -halfGrid ) {
                                        diff.x = ( ( diff.x - halfGrid ) / gridSize ) * gridSize;
                                }
                                else {
                                        diff.x = 0;
                                }

                                if ( diff.y >= halfGrid ) {
                                        diff.y = ( ( diff.y + halfGrid ) / gridSize ) * gridSize;
                                }
                                else if ( diff.y <= -halfGrid ) {
                                        diff.y = ( ( diff.y - halfGrid ) / gridSize ) * gridSize;
                                }
                                else {
                                        diff.y = 0;
                                }

                                if ( diff.z >= halfGrid ) {
                                        diff.z = ( ( diff.z + halfGrid ) / gridSize ) * gridSize;
                                }
                                else if ( diff.z <= -halfGrid ) {
                                        diff.z = ( ( diff.z - halfGrid ) / gridSize ) * gridSize;
                                }
                                else {
                                        diff.z = 0;
                                }
                        }

                        
                        //                        System.out.println( " diff=" + diff );
                        if ( diff.x != 0 || diff.y != 0 || diff.z != 0 ) {

                                
                                drag_floor_ok = true;
                                drag_ceil_ok = true;

                                // control down, drag floor only
                                if ( ( mod & evt.CTRL_MASK ) != 0 ) {
                                        drag_ceil_ok = false;
                                }
                                else if ( ( mod & evt.SHIFT_MASK ) != 0 ) {
                                        drag_floor_ok = false;
                                }
                                
                                _dragStartPt.setLocation( _dragStartPt.x + diff.x, 
                                                          _dragStartPt.y + diff.y, 
                                                          _dragStartPt.z + diff.z );
                                
                                _dragAction.diff = diff;
                                _dragAction.vertVect.clear();
                                
                                _map.doSelectionActions( _dragAction );
                                
                                for ( Iterator iter = _dragAction.vertVect.iterator(); iter.hasNext(); ) {
                                        _map.moveVertexRelative( (Map.Vertex) iter.next(), diff, true );
                                }

                                _map.notifyListenersFromCachedClip();
                        }
                }
                
                //
                // button 1 down, drag the rubberband
                //

                else if ( ( mod & evt.BUTTON1_MASK ) != 0 && _vc.isMode( _vc.MODE_SELECT ) ) {
                        
                        if ( _viewSide.getType() == _viewSide.HZ_X ) {

                                _map.setRubberBand( Math.min( _dragStartPt.x, p.x ),
                                                    - _map.getMapSize().height / 2 + 1,
                                                    Math.min( _dragStartPt.z, p.z ),
                                                    Math.abs( p.x - _dragStartPt.x ),
                                                    _map.getMapSize().height - 1,
                                                    Math.abs( p.z - _dragStartPt.z ) );
                        }
                        else {
                                _map.setRubberBand( - _map.getMapSize().height / 2 + 1,
                                                    Math.min( _dragStartPt.y, p.y ),
                                                    Math.min( _dragStartPt.z, p.z ),
                                                    _map.getMapSize().height - 1,
                                                    Math.abs( p.y - _dragStartPt.y ),
                                                    Math.abs( p.z - _dragStartPt.z ) );
                        }
                        
                        _view.setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
                        _rubberBanding = true;
                }
        }

        public void mousePressed( MouseEvent evt ) {
                Point3D p = screenToMap( evt.getPoint() );

                int mod = evt.getModifiers();
                                
                _dragStartPt.setLocation( p );

                if ( _vc.isMode( _vc.MODE_SELECT ) && ( mod & evt.BUTTON2_MASK ) != 0 ) {
                        if ( _map.getLastSelection() != null )
                                _view.setCursor( Cursor.getPredefinedCursor( Cursor.MOVE_CURSOR ) );
                        return;
                }
                
                if ( _vc.isMode( _vc.MODE_NEW_LIGHT ) ) {
                        Point3D snapped = _vc.isSnapEnabled() ? _vc.getSnapped( p ) : p;

                        if ( _map.findLight( snapped ) != null ) {
                                Toolkit.getDefaultToolkit().beep();
                                return;
                        }

                        Map.LightSource light = new Map.LightSource();
                        light.pos.setLocation( snapped.x,
                                               snapped.y,
                                               snapped.z );
                        light.r = 255;
                        light.g = 255;
                        light.b = 255;
                        light.intensity = 256;

                        _map.addLight( light );
                        return;
                }

        }
        
        public void mouseReleased( MouseEvent evt ) {

                int mod = evt.getModifiers();

                if ( _rubberBanding ) {
                        
                        final Rectangle rub_rect = 
                                new Rectangle( _map._rubberBand[0].x,
                                               _map._rubberBand[0].y,
                                               _map._rubberBand[1].x,
                                               _map._rubberBand[1].y );
                        final Rectangle big_rect =
                                new Rectangle( ( _viewSide.getType() == _viewSide.HZ_X ) ? 
                                               _map._rubberBand[0].x :
                                               _map._rubberBand[0].y,
                                               _map._rubberBand[0].z,
                                               ( _viewSide.getType() == _viewSide.HZ_X ) ? 
                                               _map._rubberBand[1].x :
                                               _map._rubberBand[1].y,
                                               _map._rubberBand[1].z );
                        
                        final Rectangle user_clip = 
                                new Rectangle( _viewSide.getType() == _viewSide.HZ_X ? 
                                               _map._clip.x : _map._clip.y,
                                               - _map.getMapHeightZ() / 2,
                                               _viewSide.getType() == _viewSide.HZ_X ? 
                                               _map._clip.width : _map._clip.height,
                                               _map.getMapHeightZ() );
                        
                        final Rectangle rect = big_rect.intersection( user_clip );

                        final int extent_x = _map._clip.x + _map._clip.width;
                        final int extent_y = _map._clip.y + _map._clip.height;

                        _map.unsetRubberBand();
                        _rubberBanding = false;
                        _view.setCursor( null );
                        
                        if ( ! evt.isControlDown() && ! evt.isShiftDown() ) {
                                _map.clearSelections( true );
                        }
                        
                        Map.Element elem = _map.getLastSelection();

                        //
                        // select all sects in rubber band area
                        //

                        if ( ( elem == null && _vc.isAutoSelect( _vc.AUTO_SECTORS ) ) ||
                             elem instanceof Map.Sector ) {
                                
                                for ( Iterator iter = _map.sectors.iterator(); iter.hasNext(); ) {
                                        
                                        Map.Sector s = (Map.Sector) iter.next();
                                        
                                        Rectangle s_rect = s.getBounds( 0 );

                                        if ( _viewSide.getType() == _viewSide.HZ_X ) {
                                                if ( s_rect.y > extent_y ||
                                                     ( s_rect.y + s_rect.height ) < _map._clip.y )
                                                        continue;
                                        }
                                        else {
                                                if ( s_rect.x > extent_x ||
                                                     ( s_rect.x + s_rect.width ) < _map._clip.x )
                                                        continue;   
 
                                                s_rect.x = s_rect.y;
                                                s_rect.width = s_rect.height;
                                        }

                                        s_rect.y = s.floorHeight;
                                        s_rect.height = s.ceilingHeight - s.floorHeight;
                                        
                                        if ( s_rect.intersects( rect ) ) {
                                                if ( evt.isControlDown() ) {
                                                        _map.toggleSelection( s, true );
                                                }
                                                else {
                                                        _map.addSelection( s, true  );
                                                }
                                        }
                                }
                        }
                        
                        //
                        // select all lights in rubber band area
                        //

                        else if ( ( elem == null && _vc.isAutoSelect( _vc.AUTO_LIGHTS ) ) ||
                                  elem instanceof Map.LightSource ) {
                                
                                rect.grow( 16, 16 );
                                
                                for ( Iterator iter = _map.lights.iterator(); iter.hasNext(); ) {
                                        Map.LightSource light = (Map.LightSource) iter.next();
                                        
                                        if ( _viewSide.getType() == _viewSide.HZ_X ) {
                                                if ( light.pos.y < _map._clip.y || light.pos.y > extent_y ) {
                                                        continue;
                                                }
                                        }
                                        else if ( light.pos.x < _map._clip.x || light.pos.x > extent_x ) {
                                                continue;
                                        }  

                                        if ( rect.contains( _viewSide.isType( _viewSide.HZ_X ) ? 
                                                            light.pos.x : 
                                                            light.pos.y, 
                                                            light.pos.z ) ) {
                                                if ( evt.isControlDown() ) {
                                                        _map.toggleSelection( light, true  );
                                                }
                                                else {
                                                        _map.addSelection( light, true  );
                                                }    
                                        }
                                }
                        }

                        //
                        // select all entities in rubber band area
                        //

                        else if ( ( elem == null &&_vc.isAutoSelect( _vc.AUTO_ENTITIES ) ) ) {
                                
                        }

                        // see if any selections on the map actually changed

                        if ( _map._isClipCached ) {
                              
                                _map.addCacheClip( rub_rect );
                                _map.notifyListenersFromCachedClip();
                        }
                        else {
                                _map.notifyListenersTransient();
                        }

                }
                
                else if ( _vc.isMode( _vc.MODE_SELECT ) && ( mod & evt.BUTTON2_MASK ) != 0 ) {
                        _view.setCursor( null );
                }
        }

        public void mouseClicked( MouseEvent evt ) {
                final Object o = evt.getSource();
                
                if ( o instanceof Component ) {
                        ( (Component) o ).requestFocus();
                }

                final Point3D p = screenToMap( evt.getPoint() );

                final int mod = evt.getModifiers();

                final boolean control_down = ( mod & evt.CTRL_MASK ) != 0;
                final boolean shift_down   = ( mod & evt.SHIFT_MASK ) != 0;
                final boolean single_select = control_down == false && shift_down == false;

                final int vtype = _viewSide.getType();
                
                final int extent_x = _map._clip.x + _map._clip.width;
                final int extent_y = _map._clip.y + _map._clip.height;
                
                if ( _vc.isMode( _vc.MODE_SELECT ) && ( mod & evt.BUTTON1_MASK ) != 0 ) {
                        
                        final Point target = new Point( vtype == _viewSide.HZ_X ? p.x : p.y, p.z );
                        
                        // make sure we aren't clicking outside the user defined clip area
                        
                        if ( vtype == _viewSide.HZ_X ) {
                                if ( p.x < _map._clip.x || p.x > ( _map._clip.x + _map._clip.width ) )
                                        return;
                        }
                        else if  ( p.y < _map._clip.y || p.y > ( _map._clip.y + _map._clip.height ) ) {
                                return;
                        }

                        Map.Element elem = null;

                        //
                        // look for appropriate lights to select
                        //

                        elem = _map.getLastSelection();

                        if ( single_select ||
                             elem == null || elem instanceof Map.LightSource ) {
                                
                                for ( Iterator iter = _map.lights.iterator(); iter.hasNext(); ) {
                                
                                        Map.LightSource light = (Map.LightSource) iter.next();
                                        
                                        if ( vtype == _viewSide.HZ_X ) {
                                                if ( light.pos.y < _map._clip.y || light.pos.y > extent_y ) {
                                                        continue;
                                                }
                                        }
                                        else if ( light.pos.x < _map._clip.x || light.pos.x > extent_x ) {
                                                continue;
                                        }  
                                        
                                        if ( VMath.inRange( vtype == _viewSide.HZ_X ? 
                                                            light.pos.x : 
                                                            light.pos.y,
                                                            light.pos.z, 
                                                            target.x, 
                                                            target.y, 
                                                            16 ) ) {
                                                
                                                if ( control_down ) {
                                                        _map.toggleSelection( light, false );
                                                }
                                                else if ( shift_down ) {
                                                        _map.addSelection( light, false );
                                                }
                                                else {
                                                        _map.setSelection( light, false );
                                                }
                                                return;
                                        }
                                }
                        }
                        
                        //
                        // look for appropriate sectors to select
                        //

                        elem = _map.getLastSelection();
                        
                        if ( single_select ||
                             elem == null || elem instanceof Map.Sector ) {
                                
                                for ( Iterator iter = _map.sectors.iterator(); iter.hasNext(); ) {
                                
                                        Map.Sector s = (Map.Sector) iter.next();
                                        
                                        Rectangle s_rect = s.getBounds( 0 );
                                        
                                        if ( vtype == _viewSide.HZ_X ) {
                                                if ( s_rect.y > extent_y ||
                                                     ( s_rect.y + s_rect.height ) < _map._clip.y )
                                                        continue;
                                        }
                                        else {
                                                if ( s_rect.x > extent_x ||
                                                     ( s_rect.x + s_rect.width ) < _map._clip.x )
                                                        continue;
                                                
                                                s_rect.x = s_rect.y;
                                                s_rect.width = s_rect.height;
                                        }

                                        s_rect.y = s.floorHeight;
                                        s_rect.height = s.ceilingHeight - s.floorHeight;
                                        
                                        if ( s_rect.contains( target ) ) {
                                        
                                                if ( control_down ) {
                                                        _map.toggleSelection( s, false );
                                                }
                                                else if ( shift_down ) {
                                                        _map.addSelection( s, false );
                                                }
                                                else {
                                                        _map.setSelection( s, false );
                                                }
                                                return;
                                        }
                                }
                        }

                        if ( ! shift_down && ! control_down )
                                _map.clearSelections( false );
                        
                        return;
                }
        }
}
