/*
 * EditPanel.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;


public class EditPanel extends JPanel implements Map.Listener {
        
        MapPanel    _mapPanel = new MapPanel();

        LinePanel   _linePanel = new LinePanel();
        SectorPanel _sectorPanel = new SectorPanel();
        VertexPanel _vertexPanel = new VertexPanel();
        LightPanel  _lightPanel = new LightPanel();
        EntityPanel _entityPanel = new EntityPanel();
        
        JPanel _curPanel = null;
        
        Map _map = null;
        
        JPanel _emptyPanel = new JPanel();
        
    //        JTabbedPane _tabbed = new JTabbedPane();

        CardLayout _card  = new CardLayout();

        boolean _isSettingElement = false;

        EditPanel( Map map ) {
                
                setLayout( _card );
               
                add( _mapPanel,    "Map" );
                add( _sectorPanel, "Sector" );
                add( _linePanel,   "Line" );
                add( _vertexPanel, "Vertex" );
                add( _lightPanel,  "Light" );
                add( _entityPanel, "Entity" );
                
                setMap( map );

        }
        
        void setMap( Map map ) {
                if ( _map != null )
                        _map.listeners.removeElement( this );
                
                _map = map;
                
                if ( _map != null )
                        _map.listeners.addElement( this );
                
                setElement( null );
        }

        void setElement( Map.Element elem ) {

                if ( elem == null ) {
                        _mapPanel.setMap( _map );
                        _card.show( this, "Map" );
                        return;
                }
                
                _isSettingElement = true;
                
                if ( elem instanceof Map.Sector ) {
                        _sectorPanel.setSector( (Map.Sector) elem );
                        _card.show( this, "Sector" );
                }
                else if ( elem instanceof Map.Line ) {
                        _linePanel.setLine( (Map.Line) elem );
                        _card.show( this, "Line" );
                }
                else if ( elem instanceof Map.Vertex ) {
                        _vertexPanel.setVertex( (Map.Vertex) elem );
                        _card.show( this, "Vertex" );
                }
                else if ( elem instanceof Map.LightSource ) {
                        _lightPanel.setLight( (Map.LightSource) elem );
                        _card.show( this, "Light" );
                }
                else if ( elem instanceof Map.Entity ) {
                        _entityPanel.setEntity( (Map.Entity) elem );
                        _card.show( this, "Entity" );
                }
                
                _isSettingElement = false;

        }

        public void mapChanged( Rectangle clip ) {
                setElement( _map.getLastSelection() );
        }

        public void mapTransientChanged() {}

        public void mapSizeChanged() {
        }

        class VertexPanel extends JPanel {

                JTextField _xField = new JTextField( 5 );
                JTextField _yField = new JTextField( 5 );
                
                Map.Vertex _vert = null;
                
                public VertexPanel() {
                        _xField.setEditable( false );
                        _yField.setEditable( false );
                        
                        _xField.setHorizontalAlignment( JTextField.RIGHT );
                        _yField.setHorizontalAlignment( JTextField.RIGHT );

                        setLayout( new GridBagLayout() );
                        
                        GridBagConstraints gbc = new GridBagConstraints();
                        
                        gbc.gridx   = gbc.gridy = 0;
                        gbc.weightx = 1;
                        gbc.weighty = 0;
                        gbc.insets  = new Insets( 2, 2, 2, 2 );

                        gbc.gridwidth = 2;
                        add( new JLabel( "Vertex Info" ), gbc );

                        gbc.gridwidth = 1;
                        gbc.anchor = gbc.NORTHWEST;
                        ++gbc.gridy;
                        add( new JLabel( "X" ), gbc );
                        ++gbc.gridx;
                        add( _xField, gbc );
                        
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Y" ), gbc );
                        ++gbc.gridx;
                        add( _yField, gbc );

                        gbc.weighty = 1;
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        gbc.fill = gbc.BOTH;

                        add( new JPanel(), gbc );
                }

                public void setVertex( Map.Vertex vert ) {
                        _vert = vert;

                        _xField.setText( Integer.toString( vert.x ) );
                        _yField.setText( Integer.toString( vert.y) );
                }
        }

        class LightPanel extends JPanel {
                
                JTextField[] _posFields = new JTextField[] {
                        new JTextField( 5 ),
                        new JTextField( 5 ),
                        new JTextField( 5 )
                                };
                
                JTextField _intensityField = new JTextField( 5 );
                JButton _colorButton = new JButton( "Choose" );
                
                Map.LightSource _light = null;
                
                MultiSlider _multiSlider = new MultiSlider( 3,
                                                            new String[] { "R", "G", "B" },
                                                            new int[] { 0, 0, 0 },
                                                            new int[] { 255, 255, 255 },
                                                            new boolean[] { true, true, true },
                                                            new boolean[] { true, true, true } );

                MultiSlider.Listener _multiSliderListener = new MultiSlider.Listener() {
                                public void changed( int which, int value ) {

                                        if ( _isSettingElement )
                                                return;

                                        _update.updateLights( _update.COLOR, 
                                                              _multiSlider.getValueAt( 0 ),
                                                              _multiSlider.getValueAt( 1 ),
                                                              _multiSlider.getValueAt( 2 ) );

                                        _colorButton.setBackground( new Color( _light.r, _light.g, _light.b ) );
                                        _colorButton.setForeground( new Color( ~_light.r & 0xff,
                                                                               ~_light.g & 0xff,
                                                                               ~_light.b & 0xff ) );
                                }

                        };

                class UpdateAction {
                        
                        public final int INTENSITY = 1;
                        public final int COLOR = 2;
                        
                        private int _state = 0;
                        
                        public boolean inProgress = false;
                        
                        private int[] _args = new int[3];
                        
                        private Map.ElementAction theAction = new Map.ElementAction() {
                                        private void myAction( Map.LightSource light ) {
                                                
                                                switch ( _state ) {
                                                case INTENSITY:
                                                        light.setIntensity( _args[0] );
                                                        break;
                                                case COLOR:
                                                        light.setColor( _args[0], _args[1], _args[2] );
                                                        break;
                                                }
                                        }
                                        
                                        public void doAction( Map.Element elem ) {
                                                if ( elem instanceof Map.LightSource ) {
                                                        myAction( (Map.LightSource) elem );
                                                }
                                        }
                                };
                        
                        public synchronized void updateLights( int state, int arg1, int arg2, int arg3 ) {
                                inProgress = true;
                                _state = state;
                                _args[0] = arg1;
                                _args[1] = arg2;
                                _args[2] = arg3;
                                _map.doSelectionActions( this.theAction );
                                inProgress = false;
                        }
                };
                
                private UpdateAction _update = new UpdateAction();
                
                private ActionListener _colorButtonListener = new ActionListener() {
                                public void actionPerformed( ActionEvent evt ) {
                                        
                                        if ( _isSettingElement )
                                                return;

                                        Color c = JColorChooser.showDialog( LightPanel.this,
                                                                            "Choose Light Source Color",
                                                                            _colorButton.getBackground() );
                                        
                                        if ( c == null || c.equals( _colorButton.getBackground() ) )
                                                return;
                                        
                                        _update.updateLights( _update.COLOR, c.getRed(), c.getGreen(), c.getBlue() );

                                        _multiSliderListener.setListening( false );
                                        _multiSlider.setValueAt( 0, _light.r );
                                        _multiSlider.setValueAt( 1, _light.g );
                                        _multiSlider.setValueAt( 2, _light.b );
                                        _multiSliderListener.setListening( true );

                                        _colorButton.setBackground( c );
                                        _colorButton.setForeground( new Color( ~_light.r & 0xff, ~_light.g & 0xff, ~_light.b & 0xff ) );     
                                }
                        };

                public LightPanel() {
                        
                        for ( int i = 0; i < _posFields.length; ++i ) {
                                _posFields[i].setEditable( false );
                                _posFields[i].setHorizontalAlignment( JTextField.RIGHT );
                        }
                        
                        setLayout( new GridBagLayout() );
                        
                        GridBagConstraints gbc = new GridBagConstraints();
                        
                        gbc.gridx = gbc.gridy = 0;
                        gbc.weightx = 1;
                        gbc.weighty = 0;
                        gbc.insets = new Insets( 2, 2, 2, 2 );

                        gbc.gridwidth = 2;
                        add( new JLabel( "Light Info" ), gbc );
                        
                        gbc.gridwidth = 1;
                        ++gbc.gridy;
                        gbc.fill = gbc.BOTH;
                        gbc.anchor = gbc.NORTHEAST;
                        add( new JLabel( "X" ), gbc );
                        ++gbc.gridx;
                        add( _posFields[0], gbc );
                        
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Y" ), gbc );
                        ++gbc.gridx;
                        add( _posFields[1], gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;

                        add( new JLabel( "Z" ), gbc );
                        ++gbc.gridx;
                        add( _posFields[2], gbc );

                        _intensityField.addActionListener( _fieldListener );
                        
                        FocusHelper helper = new FocusHelper( _fieldListener );

                        _intensityField.addFocusListener( helper );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Intensity" ), gbc );
                        ++gbc.gridx;
                        add( _intensityField, gbc );

                        _colorButton.addActionListener( _colorButtonListener );

                        JPanel colorPanel = new JPanel( new BorderLayout() );
                        colorPanel.setBorder( BorderFactory.createTitledBorder( "Color" ) );
                        colorPanel.add( _multiSlider, "Center" );
                        colorPanel.add( _colorButton, "South" );
                        
                        _multiSlider.addListener( _multiSliderListener );
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        gbc.gridwidth = 2;
                        gbc.fill = gbc.BOTH;
                        add( colorPanel, gbc );
                        
                        gbc.weighty = 1;
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        gbc.fill = gbc.BOTH;

                        add( new JPanel(), gbc );
                }

                public void setLight( Map.LightSource light ) {
                        _light = light;
                        
                        _posFields[0].setText( Integer.toString( light.pos.x ) );
                        _posFields[1].setText( Integer.toString( light.pos.y ) );
                        _posFields[2].setText( Integer.toString( light.pos.z ) );

                        _multiSliderListener.setListening( false );
                        _multiSlider.setValueAt( 0, _light.r );
                        _multiSlider.setValueAt( 1, _light.g );
                        _multiSlider.setValueAt( 2, _light.b );
                        _multiSliderListener.setListening( true );

                        _intensityField.setText( "" + _light.intensity );
                        _colorButton.setBackground( new Color( _light.r, _light.g, _light.b ) );
                        _colorButton.setForeground( new Color( ~_light.r & 0xff, ~_light.g & 0xff, ~_light.b & 0xff ) );

                }

                private ActionListener _fieldListener = new ActionListener() {
                                public void actionPerformed( ActionEvent evt ) {
                                        
                                        if ( _isSettingElement )
                                                return;

                                        try {
                                                int val = Integer.parseInt( _intensityField.getText() );
                                                int truval = Math.min( Math.max( 0, val ), 4096 );
                                                
                                                _update.updateLights( _update.INTENSITY, truval, 0, 0 );
                                                //                                                _light.intensity = truval;

                                                if ( val != truval )
                                                        _intensityField.setText( "" + truval );
                                        }
                                        catch ( NumberFormatException exp ) {
                                                Toolkit.getDefaultToolkit().beep();
                                        }
                                }
                        };
        }

        class EntityPanel extends JPanel {
                
                JTextField[] _posFields = new JTextField[] {
                        new JTextField( 5 ),
                        new JTextField( 5 ),
                        new JTextField( 5 )
                                };
                
                Map.Entity _entity = null;
                
                public EntityPanel() {
                        
                        for ( int i = 0; i < _posFields.length; ++i ) {
                                _posFields[i].setEditable( false );
                                _posFields[i].setHorizontalAlignment( JTextField.RIGHT );
                        }
                        
                        setLayout( new GridBagLayout() );
                        
                        GridBagConstraints gbc = new GridBagConstraints();
                        
                        gbc.gridx = gbc.gridy = 0;
                        gbc.weightx = 1;
                        gbc.weighty = 0;

                        add( new JLabel( "X" ), gbc );
                        ++gbc.gridx;
                        add( _posFields[0], gbc );
                        
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Y" ), gbc );
                        ++gbc.gridx;
                        add( _posFields[1], gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Z" ), gbc );
                        ++gbc.gridx;
                        add( _posFields[2], gbc );

                        gbc.weighty = 1;
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        gbc.fill = gbc.BOTH;

                        add( new JPanel(), gbc );
                }

                public void setEntity( Map.Entity ent ) {
                        _entity = ent;
                        
                        _posFields[0].setText( Integer.toString( _entity.pos.x ) );
                        _posFields[1].setText( Integer.toString( _entity.pos.y ) );
                        _posFields[2].setText( Integer.toString( _entity.pos.z ) );
                }
        }

        class MapPanel extends JPanel implements Map.Listener {

                private JTextField _nameField = new JTextField( 10 );
                private JPanel     _fill = new JPanel();
                private JComboBox  _defaultWallTexCombo    = new JComboBox( TextureManager.getTexnames() );
                private JComboBox  _defaultFloorTexCombo   = new JComboBox( TextureManager.getTexnames() );
                private JComboBox  _defaultCeilingTexCombo = new JComboBox( TextureManager.getTexnames() );
                
                private JTextField _numSectsField = new JTextField( 6 );
                private JTextField _numLinesField = new JTextField( 6 );
                private JTextField _numVertsField = new JTextField( 6 );
                
                private JTextField _mapWidthField = new JTextField( 6 );
                private JTextField _mapHeightField = new JTextField( 6 );

                MapPanel() {

                        _defaultWallTexCombo.setMaximumRowCount( 20 );
                        _defaultFloorTexCombo.setMaximumRowCount( 20 );
                        _defaultCeilingTexCombo.setMaximumRowCount( 20 );

                        _numSectsField.setEditable(  false );
                        _numLinesField.setEditable(  false );
                        _numVertsField.setEditable(  false );
                        
                        _numSectsField.setHorizontalAlignment( JTextField.RIGHT );
                        _numLinesField.setHorizontalAlignment( JTextField.RIGHT );
                        _numVertsField.setHorizontalAlignment( JTextField.RIGHT );
                        _mapWidthField.setHorizontalAlignment(  JTextField.RIGHT );
                        _mapHeightField.setHorizontalAlignment( JTextField.RIGHT );

                        _defaultFloorTexCombo.setSelectedIndex( TextureManager.findTexture( Editor.getProperty( Editor.DEFAULT_FLOOR_TEX ) ) );
                        _defaultCeilingTexCombo.setSelectedIndex( TextureManager.findTexture( Editor.getProperty( Editor.DEFAULT_CEILING_TEX ) ) );
                        _defaultWallTexCombo.setSelectedIndex( TextureManager.findTexture( Editor.getProperty( Editor.DEFAULT_WALL_TEX ) ) );

                        setLayout( new GridBagLayout() );
                        
                        GridBagConstraints gbc = new GridBagConstraints();
                        
                        gbc.gridx = gbc.gridy = 0;
                        gbc.weightx = 1;
                        gbc.weighty = 0;
                        gbc.insets = new Insets( 2, 2, 2, 2 );

                        gbc.gridwidth = 2;

                        add( new JLabel( "Map Info" ), gbc );
                        ++gbc.gridy;

                        gbc.anchor = gbc.NORTHWEST;

                        add( new JLabel( "Name:" ), gbc );
                        ++gbc.gridy;
                        add( _nameField, gbc );
                        
                        ++gbc.gridy;
                        add( new JLabel( "Size:" ), gbc );
                        ++gbc.gridy;
                        
                        gbc.gridwidth = 1;
                        
                        add( _mapWidthField, gbc );
                        ++gbc.gridx;
                        add( _mapHeightField, gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Num Sects" ), gbc );
                        ++gbc.gridx;
                        add( _numSectsField, gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Num Verts" ), gbc );
                        ++gbc.gridx;
                        add( _numVertsField, gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Num Lines" ), gbc );
                        ++gbc.gridx;
                        add( _numLinesField, gbc );

                        gbc.gridwidth = 2;
                        
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Default Floor Tex" ), gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( _defaultFloorTexCombo, gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Default Ceiling Tex" ), gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( _defaultCeilingTexCombo, gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Default Wall Tex" ), gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( _defaultWallTexCombo, gbc );

                        gbc.weighty = 1;
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        gbc.fill = gbc.BOTH;
                        add( _fill, gbc );
                        
                        KeyListener key = new KeyAdapter() {
                                        public void keyReleased( KeyEvent evt ) {
                                                Object source = evt.getSource();
                                                
                                                if ( source == _nameField ) {
                                                        _map.name = _nameField.getText();
                                                }
                                        }
                                };

                        _nameField.addKeyListener( key );

                        ActionListener listener = new ActionListener() {
                                        public void actionPerformed( ActionEvent evt ) {
                                                
                                                if ( _isSettingElement )
                                                        return;

                                                Object source = evt.getSource();
                                                
                                                if ( source == _defaultFloorTexCombo ) {
                                                        _map.setDefaultTex( _map.DEF_FLOOR_TEX, 
                                                                        _defaultFloorTexCombo.getSelectedIndex() );
                                                        Editor.setProperty( Editor.DEFAULT_FLOOR_TEX,
                                                                            (String) _defaultFloorTexCombo.getSelectedItem() );
                                                        Editor.saveProperties();
                                                }
                                                else if ( source == _defaultCeilingTexCombo ) {
                                                        _map.setDefaultTex( _map.DEF_CEILING_TEX,
                                                                        _defaultCeilingTexCombo.getSelectedIndex() );
                                                        
                                                        Editor.setProperty( Editor.DEFAULT_CEILING_TEX,
                                                                            (String) _defaultCeilingTexCombo.getSelectedItem() );
                                                        Editor.saveProperties();
                                                }
                                                else if ( source == _defaultWallTexCombo ) {
                                                        _map.setDefaultTex( _map.DEF_WALL_TEX,
                                                                     _defaultWallTexCombo.getSelectedIndex() );
                                                        Editor.setProperty( Editor.DEFAULT_WALL_TEX,
                                                                            (String) _defaultWallTexCombo.getSelectedItem() );
                                                        Editor.saveProperties();
                                                }
                                                else if ( source == _mapWidthField ) {
                                                        try {
                                                                int w = Integer.parseInt( _mapWidthField.getText() );
                                                                _map.setMapSize( w, _map.getMapSize().height );
                                                        }
                                                        catch ( Exception exp ) {
                                                                mapSizeChanged();
                                                        }
                                                }
                                                else if ( source == _mapHeightField ) {
                                                        try {
                                                                int h = Integer.parseInt( _mapHeightField.getText() );
                                                                _map.setMapSize( _map.getMapSize().width, h );
                                                        }
                                                        catch ( Exception exp ) {
                                                                mapSizeChanged();
                                                        }
                                                }
                                        }
                                };

                        _defaultFloorTexCombo.addActionListener(   listener );
                        _defaultCeilingTexCombo.addActionListener( listener );
                        _defaultWallTexCombo.addActionListener(    listener );
                        _mapHeightField.addActionListener( listener );
                        _mapWidthField.addActionListener( listener );

                        _mapHeightField.setToolTipText( "Change height and hit enter to modify" );
                        _mapWidthField.setToolTipText( "Change width and hit enter to modify" );
                        

                }

                public void mapChanged( Rectangle clip ) {
                        _numSectsField.setText( Integer.toString( _map.sectors.size() ) );
                        _numLinesField.setText( Integer.toString( _map.lines.size() ) );
                        _numVertsField.setText( Integer.toString( _map.vertexes.size() ) );
                }
                
                public void mapTransientChanged() {}
                
                public void mapSizeChanged() {
                        Dimension s = _map.getMapSize();
                        _mapWidthField.setText( Integer.toString( s.width ) );
                        _mapHeightField.setText( Integer.toString( s.height ) );
                }
                
                void setMap( Map map ) {
                        
                        if ( _map != null ) {
                                _map.listeners.removeElement( this );
                        }
                        
                        _nameField.setText( _map.name );
                        if ( _map.getDefaultTex( _map.DEF_FLOOR_TEX ) > 0 )
                                _defaultFloorTexCombo.setSelectedIndex(   _map.getDefaultTex( _map.DEF_FLOOR_TEX ) );
                        if ( _map.getDefaultTex( _map.DEF_CEILING_TEX ) > 0 )
                                _defaultCeilingTexCombo.setSelectedIndex( _map.getDefaultTex( _map.DEF_CEILING_TEX ) );
                        if ( _map.getDefaultTex( _map.DEF_WALL_TEX ) > 0 )
                                _defaultWallTexCombo.setSelectedIndex(    _map.getDefaultTex( _map.DEF_WALL_TEX ) );
                        
                        mapSizeChanged();

                        _map.listeners.addElement( this );
                        
                        mapChanged( null );
                }
        }

        class LinePanel extends JPanel {
                
                SpecialPanel _specialPanel = new SpecialPanel();
                Map.Line _line = null;

                SidePanel[] _sidePanels = new SidePanel[] {
                        new SidePanel(),
                        new SidePanel()
                                };

                JTabbedPane _lineTabbed = new JTabbedPane();
                
                LinePanel() {
                        _lineTabbed.addTab( "Front", _sidePanels[0] );
                        _lineTabbed.addTab( "Back",  _sidePanels[1] );
                        _lineTabbed.setEnabledAt( 0, false );
                        _lineTabbed.setEnabledAt( 1, false );
                        setLayout( new GridBagLayout() );

                        GridBagConstraints gbc = new GridBagConstraints();

                        gbc.insets = new Insets( 2, 2, 2, 2 );
                        gbc.weightx = 1;

                        gbc.gridx = 0;
                        gbc.gridy = 0;
                        gbc.fill    = gbc.HORIZONTAL;   
                        add( new JLabel( "Line Info" ), gbc );
                        
                        gbc.anchor = gbc.NORTHWEST;
                        ++gbc.gridy;
                        add( new JLabel( "Sides" ), gbc );
                        
                        ++gbc.gridy;
                        add( _lineTabbed, gbc );

                        ++gbc.gridy;
                        add( _specialPanel, gbc );

                        gbc.weighty = 1;
                        gbc.fill    = gbc.BOTH;
                        add( new JPanel(), gbc );
                        
                        _lineTabbed.addChangeListener( new ChangeListener() {
                                        public void stateChanged( ChangeEvent evt ) {
                                                _map.preferredLineSide = _lineTabbed.getSelectedIndex();
                                        }
                                } );
                }
                
                void setLine( Map.Line line ) {
                        _line = line;

                        if ( line.line_sides[0] != null ) {
                                _sidePanels[0].setSide( line.line_sides[0] );
                        }

                        if ( line.line_sides[1] != null ) {
                                _sidePanels[1].setSide( line.line_sides[1] );
                        }
                        
                        _lineTabbed.setEnabledAt( 0, line.line_sides[0] != null );
                        _lineTabbed.setEnabledAt( 1, line.line_sides[1] != null );
                        
                        if ( line.line_sides[1] != null && _map.preferredLineSide == 1 )
                                _lineTabbed.setSelectedIndex( 1 );
                        else
                                _lineTabbed.setSelectedIndex( 0 );

                        _specialPanel.updateSpecial();

                        invalidate();
                }

                class SpecialPanel extends JPanel {
                        JCheckBox  _specialCheck = new JCheckBox( "Is Special" );

                        JTextField _targetField  = new JTextField( 4 );

                        JCheckBox[] _checks = new JCheckBox[ Map.Line.trigger_flagnames.length ];
                        
                        JLabel[] _labels = new JLabel[] {
                                new JLabel( "Target" )
                                        };
                                
                        public SpecialPanel() {

                                setLayout( new GridBagLayout() );
                                GridBagConstraints gbc = new GridBagConstraints();

                                gbc.gridwidth = 2;
                                gbc.fill = gbc.BOTH;
                                gbc.weightx = 1;
                                gbc.gridx = gbc.gridy = 0;
                                gbc.anchor = gbc.NORTHEAST;
                        
                                setBorder( BorderFactory.createTitledBorder( "Line Special" ) );

                                int lindex = 0;

                                gbc.gridwidth = 2;
                                                                
                                add( _specialCheck, gbc );
                                ++gbc.gridy;
                                
                                gbc.gridwidth = 1;
                                ++gbc.gridy;
                                gbc.gridx = 0;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridx;
                                add( _targetField, gbc );

                                gbc.gridwidth = 2;
                                gbc.gridx = 0;
                                
                                for ( int i = 0; i <  Map.Line.trigger_flagnames.length; ++i ) {
                                        _checks[i] = new JCheckBox( Map.Line.trigger_flagnames[i] );
                                        ++gbc.gridy;
                                        add( _checks[i], gbc );
                                        _checks[i].addActionListener( _specListener );
                                }
                                
                                _specialCheck.addActionListener( _specListener );
                                _targetField.addActionListener( _specListener );

                                FocusHelper helper = new FocusHelper( _specListener );

                                _targetField.addFocusListener( helper );
                                
                        }

                        public void updateSpecial() {

                                Script script = _line == null ? null : _line._script;
                                
                                boolean state = ( script != null );

                                if ( script == null ) {
                                        _targetField.setText( "" );
                                        for ( int i = 0; i < _checks.length; ++i ) {
                                                _checks[i].setEnabled( false );
                                        }                              
                                }
                                else {
                                        _targetField.setText( "" + script.num );
                                        
                                        for ( int i = 0; i < _checks.length; ++i ) {
                                                _checks[i].setSelected( ( _line.trigger_flags & ( 1 << i ) ) != 0 );
                                                _checks[i].setEnabled( true );

                                       }
                                }

                                _specialCheck.setSelected( state );
                                _targetField.setEnabled( state );
                            
                                for ( int i = 0; i < _labels.length; ++i ) {
                                        _labels[i].setEnabled( state );
                                }
                        }

                        private Map.Line a_line = null;
                        
                        private Map.ElementAction theAction = new Map.ElementAction() {

                                        public void doAction( Map.Element elem ) {
                                                if ( elem instanceof Map.Line ) {

                                                        Map.Line line = (Map.Line) elem;

                                                        line._script = a_line._script;
                                                        line.trigger_flags = a_line.trigger_flags;
                                                        line._map.addCacheClip( line.getBounds( 0 ) );
                                                }
                                        }
                                };

                        ActionListener _specListener = new ActionListener() {
                                        public void actionPerformed( ActionEvent evt ) {

                                                if ( _isSettingElement )
                                                        return;

                                                Object source = evt.getSource();

                                                if ( source == _specialCheck ) {
                                                        if ( _specialCheck.isSelected() ) {
                                                                _line.setScript( 0 );
                                                        }
                                                        else {
                                                                _line._script = null;
                                                                _map.notifyListeners( _line.getBounds( 0 ) );
                                                        }
                                                }
                                                
                                                else if ( source == _targetField ) {
                                                        try {
                                                                int scriptnum = Byte.parseByte( _targetField.getText() );
                                                                _line.setScript( scriptnum );

                                                        } 
                                                        catch ( NumberFormatException exp ) {
                                                                System.out.println( "bad digits in line special" );
                                                                Toolkit.getDefaultToolkit().beep();
                                                                return;
                                                                // ignore
                                                        }
                                                }
                                                
                                                else if ( source instanceof JCheckBox ) {
                                                        for ( int i = 0; i < _checks.length; ++i ) {
                                                                if ( source == _checks[i] ) {
                                                                        if ( _checks[i].isSelected() )
                                                                                _line.trigger_flags |= ( 1 << i );
                                                                        else
                                                                                _line.trigger_flags &= ~( 1 << i );
                                                                }
                                                        }
                                                }

                                                a_line = _line;
                                                _map.doSelectionActions( SpecialPanel.this.theAction );
                                                _map.addCacheClip( _line.getBounds( 0 ) );
                                                        
                                                _map.notifyListenersFromCachedClip();
                                        }
                                };
                }

                class SidePanel extends JPanel {

                        JPanel _fill = new JPanel();
                        JLabel _sectNumLabel = new JLabel();
                        JComboBox[] _texCombos = new JComboBox[] {
                                new JComboBox( TextureManager.getTexnames() ),
                                new JComboBox( TextureManager.getTexnames() ),
                                new JComboBox( TextureManager.getTexnames() )
                                        };
                        Map.Side _side = null;
                        
                        final String[] textitles = new String[] {
                                "Upper Texture",
                                "Main Texture",
                                "Lower Texture"

                        };

                        private UpdateAction _update = new UpdateAction();

                        SidePanel() {
                                setLayout( new GridBagLayout() );
                                
                                GridBagConstraints gbc = new GridBagConstraints();
                                
                                gbc.gridx = gbc.gridy = 0;
                                gbc.weightx = 1;
                                gbc.weighty = 0;
                                gbc.fill = gbc.BOTH;

                                add( new JLabel( "Sect Num" ), gbc );
                                ++gbc.gridx;
                                add( _sectNumLabel, gbc );
                                
                                gbc.gridwidth = 2;

                                for ( int i = 0; i < 3; ++i ) {
                                        ++gbc.gridy;
                                        gbc.gridx = 0;
                                        add( new JLabel( textitles[i] ), gbc );
                                        
                                        ++gbc.gridy;
                                        gbc.gridx = 0;
                                        _texCombos[i].setMaximumRowCount( 20 );
                                        add( _texCombos[i], gbc );
                                }
                                
                                gbc.weighty = 1;
                                ++gbc.gridy;
                                gbc.gridx = 0;
                                gbc.fill = gbc.BOTH;
                                add( _fill, gbc );
                                
                                ActionListener listener = new ActionListener() {
                                                public void actionPerformed( ActionEvent evt ) {

                                                        if ( _isSettingElement )
                                                                return;
                                                
                                                        Object source = evt.getSource();
                                                        
                                                        if ( source == _texCombos[0] )
                                                                _update.updateLines( _update.UPPER_TEXTURE );
                                                        else if ( source == _texCombos[1] )
                                                                _update.updateLines( _update.MAIN_TEXTURE );
                                                        else if ( source == _texCombos[2] )
                                                                _update.updateLines( _update.LOWER_TEXTURE );
                                                }
                                        };
                                
                                for ( int i = 0; i < 3; ++i ) {
                                        _texCombos[i].addActionListener( listener );
                                }
                                
                        }
                
                        class UpdateAction {
                        
                                public final int UPPER_TEXTURE   = 1;
                                public final int MAIN_TEXTURE    = 2;
                                public final int LOWER_TEXTURE   = 3;

                                private int _state = 0;
                        
                                public boolean inProgress = false;
                                
                                private Map.ElementAction theAction = new Map.ElementAction() {
                                                private void myAction( Map.Line line ) {
                                                
                                                        int which = _lineTabbed.getSelectedIndex();
                                                   
                                                        Map.Side side = line.line_sides[ which ];

                                                        if ( side == null )
                                                                return;
                                                
                                                        switch ( _state ) {
                                                        case UPPER_TEXTURE:
                                                                System.out.println( " setting side " + side + " upper tex to " + 
                                                                                    _texCombos[0].getSelectedIndex() );

                                                                side.setTexture( side.UPPER_INDEX, _texCombos[0].getSelectedIndex() );
                                                                break;
                                                        case MAIN_TEXTURE:
                                                                side.setTexture( side.MAIN_INDEX,  _texCombos[1].getSelectedIndex() );
                                                                break;
                                                        case LOWER_TEXTURE:
                                                                side.setTexture( side.LOWER_INDEX, _texCombos[2].getSelectedIndex() );
                                                                break;
                                                        }
                                                }
                                        
                                                public void doAction( Map.Element elem ) {
                                                        if ( elem instanceof Map.Line ) {
                                                                myAction( (Map.Line) elem );
                                                        }
                                                }
                                        };
                        
                                public synchronized void updateLines( int state ) {
                                        inProgress = true;
                                        _state = state;
                                        _map.doSelectionActions( this.theAction );
                                        inProgress = false;
                                }
                        };

                        void setSide( Map.Side side ) {
                                
                                //
                                // if an update is in progress and the selection is not changing,
                                // ignore this reset
                                //
                                
                                if ( side == _side && _update.inProgress )
                                        return;
                                
                                _side = side;
                                _sectNumLabel.setText( Integer.toString( _side.sect.num ) );
                                
                                Map.Side other = null;

                                other = _side.getOpposite();
                                
                                for ( int i = 0; i < 3; ++i ) {
                                        _texCombos[i].setSelectedIndex( _side.getTexture( i ) );
                                        _texCombos[i].setEnabled( _side.hasSection( i, other ) );
                                }
                        } 
                }
        }

        class SectorPanel extends JPanel {

                JPanel _fill = new JPanel();
                JLabel _numLabel = new JLabel();
                JTextField _floorField = new JTextField( 4 );
                JTextField _ceilField = new JTextField( 4 );
                JComboBox _floorTexCombo = new JComboBox( TextureManager.getTexnames() );
                JComboBox _ceilTexCombo  = new JComboBox( TextureManager.getTexnames() );
                
                SpecialPanel _specialPanel    = new SpecialPanel();

                Map.Sector _sect = null;
                
                private UpdateAction _update = new UpdateAction();

                SectorPanel() {

                        _floorTexCombo.setMaximumRowCount( 20 );
                        _ceilTexCombo.setMaximumRowCount( 20 );

                        setLayout( new GridBagLayout() );
                        
                        GridBagConstraints gbc = new GridBagConstraints();

                        gbc.gridx = gbc.gridy = 0;
                        gbc.weightx = 1;
                        gbc.weighty = 0;
                        gbc.insets = new Insets( 2, 2, 2, 2 );

                        gbc.gridwidth = 2;
                        add( new JLabel( "Sector Info" ), gbc );
                        gbc.gridwidth = 1;
                        ++gbc.gridy;

                        gbc.fill = gbc.HORIZONTAL;
                        add( new JLabel( "Num" ), gbc );
                        ++gbc.gridx;
                        add( _numLabel, gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Floor" ), gbc );
                        ++gbc.gridx;
                        add( _floorField, gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Ceil" ), gbc );
                        ++gbc.gridx;
                        add( _ceilField, gbc );
                        
                        gbc.gridwidth = 2;
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Floor Texture" ), gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( _floorTexCombo, gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( new JLabel( "Ceiling Texture" ), gbc );

                        ++gbc.gridy;
                        gbc.gridx = 0;
                        add( _ceilTexCombo, gbc );

                        gbc.weighty = 0;
                        ++gbc.gridy;
                        gbc.gridx = 0;
                        gbc.anchor = gbc.NORTHEAST;
                        
                        add( _specialPanel, gbc );
                        
                        gbc.weighty = 1;
                        gbc.fill = gbc.BOTH;
                        add( _fill, gbc );

                        ActionListener listener = new ActionListener() {
                                        public void actionPerformed( ActionEvent evt ) {

                                                if ( _isSettingElement )
                                                        return;
                                                
                                                Object source = evt.getSource();
                                                
                                                try {
                                                        if ( source == _floorTexCombo ) {
                                                                _update.updateSects( _update.FLOOR_TEXTURE,
                                                                                     _floorTexCombo.getSelectedIndex() );
                                                        }
                                                        else if ( source == _ceilTexCombo ) {
                                                                _update.updateSects( _update.CEILING_TEXTURE,
                                                                                     _ceilTexCombo.getSelectedIndex() );
                                                        }
                                                        else if ( source == _ceilField ) {
                                                                String text = _ceilField.getText();
                                                                _update.updateSects( _update.CEILING_HEIGHT,
                                                                                     Integer.parseInt( text ) );
                                                        }
                                                        else if ( source == _floorField ) {
                                                                String text = _floorField.getText();
                                                                _update.updateSects( _update.FLOOR_HEIGHT,
                                                                                     Integer.parseInt( text ) );
                                                        }
                                                }
                                                catch ( NumberFormatException exp ) {} // not an error
                                        }
                                };

                        _floorTexCombo.addActionListener( listener );
                        _ceilTexCombo.addActionListener( listener );

                        _floorField.addActionListener( listener );
                        _ceilField.addActionListener( listener );

                        FocusHelper helper = new FocusHelper( listener );

                        _floorField.addFocusListener( helper );
                        _ceilField.addFocusListener( helper );

                        _specialPanel.updateSpecial();

                }


                class UpdateAction {
                        
                        public final int FLOOR_HEIGHT    = 1;
                        public final int CEILING_HEIGHT  = 2;
                        public final int FLOOR_TEXTURE   = 3;
                        public final int CEILING_TEXTURE = 4;
                        public final int LIGHTING        = 5;

                        private int _state = 0;
                        private int _value = 0;
                        public boolean inProgress = false;
                        
                        private Map.ElementAction theAction = new Map.ElementAction() {
                                        private void myAction( Map.Sector sect ) {
                                                switch ( _state ) {
                                                case FLOOR_HEIGHT:
                                                        sect.setFloorHeight( _value, true );
                                                        break;
                                                case CEILING_HEIGHT:
                                                        sect.setCeilingHeight( _value, true );
                                                        break;
                                                case FLOOR_TEXTURE:
                                                        sect.setTex( sect.FLOOR_TEX, _value );
                                                        break;
                                                case CEILING_TEXTURE:
                                                        sect.setTex( sect.CEILING_TEX, _value );
                                                        break;
                                                }
                                        }
                                        
                                        public void doAction( Map.Element elem ) {
                                                if ( elem instanceof Map.Sector ) {
                                                        myAction( (Map.Sector) elem );
                                                }
                                        }
                                };
                        
                        public synchronized void updateSects( int state, int value ) {
                                inProgress = true;
                                _state = state;
                                _value = value;
                                _map.doSelectionActions( this.theAction );
                                inProgress = false;
                                _map.notifyListenersFromCachedClip();
                        }
                };
                
                
                void setSector( Map.Sector s ) {

                        //
                        // if an update is in progress and the selection is not changing,
                        // ignore this reset
                        //

                        if ( s == _sect && _update.inProgress )
                                return;

                        _sect = s;
                        _numLabel.setText( Integer.toString( s.num ) );
                        _floorField.setText( Integer.toString( s.floorHeight ) );
                        _ceilField.setText( Integer.toString( s.ceilingHeight ) );
                        _floorTexCombo.setSelectedIndex( s.getTexIndex( s.FLOOR_TEX ) );
                        _ceilTexCombo.setSelectedIndex( s.getTexIndex( s.CEILING_TEX ) );

                        invalidate();
                        
                        _specialPanel.updateSpecial();
                }

                class SpecialPanel extends JPanel {
                        JCheckBox  _specialCheck = new JCheckBox( "Is Special" );
                        JComboBox  _specialCombo = new JComboBox( Map.SectSpecial.names );
                        JComboBox  _soundCombo   = new JComboBox( Map.SectSpecial.sndNames );

                        JTextField _travelField  = new JTextField( 4 );
                        JTextField _tagField     = new JTextField( 4 );
                        JTextField _delayField   = new JTextField( 4 );
                        JTextField _speedField   = new JTextField( 4 );
                        JTextField _damageField  = new JTextField( 4 );
                        JTextField _iterField    = new JTextField( 4 );

                        JCheckBox  _startCheck   = new JCheckBox( "Starts Extended" );

                        JLabel[] _labels = new JLabel[] {
                                new JLabel( "Type" ),
                                new JLabel( "Sounds" ),
                                new JLabel( "Tag" ),
                                new JLabel( "Travel" ),
                                new JLabel( "Speed" ),
                                new JLabel( "Delay" ),
                                new JLabel( "Damage" ),
                                new JLabel( "Iterations" )
                                        };
                                
                        public SpecialPanel() {
                                setLayout( new GridBagLayout() );
                                GridBagConstraints gbc = new GridBagConstraints();

                                int lindex = 0;
                                
                                gbc.gridwidth = 2;
                                gbc.fill = gbc.BOTH;
                                gbc.weightx = 1;
                                gbc.gridx = gbc.gridy = 0;
                                gbc.anchor = gbc.NORTHEAST;

                                setBorder( BorderFactory.createTitledBorder( "Sector Special" ) );
                                
                                add( _specialCheck, gbc );
                                ++gbc.gridy;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridy;
                                add( _specialCombo, gbc );
                                ++gbc.gridy;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridy;
                                add( _soundCombo, gbc );

                                gbc.gridwidth = 1;
                        
                                ++gbc.gridy;
                                gbc.gridx = 0;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridx;
                                add( _tagField, gbc );

                                ++gbc.gridy;
                                gbc.gridx = 0;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridx;
                                add( _travelField, gbc );

                                ++gbc.gridy;
                                gbc.gridx = 0;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridx;
                                add( _speedField, gbc );

                                ++gbc.gridy;
                                gbc.gridx = 0;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridx;
                                add( _delayField, gbc );

                                ++gbc.gridy;
                                gbc.gridx = 0;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridx;
                                add( _damageField, gbc );

                                ++gbc.gridy;
                                gbc.gridx = 0;
                                add( _labels[ lindex++ ], gbc );
                                ++gbc.gridx;
                                add( _iterField, gbc );

                                ++gbc.gridy;
                                gbc.gridx = 0;
                                gbc.gridwidth = 2;
                                add( _startCheck, gbc );


                                _specialCheck.addActionListener( _specListener );
                                _specialCombo.addActionListener( _specListener );
                                _soundCombo.addActionListener(   _specListener );
                                _travelField.addActionListener(  _specListener );
                                _tagField.addActionListener(     _specListener );
                                _speedField.addActionListener(   _specListener );
                                _damageField.addActionListener(  _specListener );
                                _delayField.addActionListener(   _specListener );
                                _iterField.addActionListener(    _specListener );
                                _startCheck.addActionListener(   _specListener );

                                FocusHelper helper = new FocusHelper( _specListener );
                                
                                _travelField.addFocusListener(  helper );
                                _tagField.addFocusListener(     helper );
                                _speedField.addFocusListener(   helper );
                                _damageField.addFocusListener(  helper );
                                _delayField.addFocusListener(   helper );
                                _iterField.addFocusListener(    helper );

                        }

                        public void updateSpecial() {
                                Map.SectSpecial spec = _sect == null ? null : _sect._special;
                                
                                boolean state = ( spec != null );
                                
                                if ( spec == null ) {
                                        _tagField.setText( "" );
                                        _travelField.setText( "" );
                                        _speedField.setText( "" );
                                        _damageField.setText( "" );
                                        _delayField.setText( "" );
                                        _iterField.setText( "" );
                               }
                                else {
                                        _tagField.setText( "" + spec._tag );
                                        _travelField.setText( "" + spec._travel );
                                        _speedField.setText( "" + spec._speed );
                                        _damageField.setText( "" + spec._damage );
                                        _delayField.setText( "" + spec._delay );
                                        _iterField.setText( "" + spec._iterations );
                                        _startCheck.setSelected( spec._startExtended );
                                        _specialCombo.setSelectedIndex( spec._type );
                                        _soundCombo.setSelectedIndex( spec._sound );
                                }
                                
                                _specialCheck.setSelected( state );
                                _delayField.setEnabled(    state );
                                _specialCombo.setEnabled(  state );
                                _soundCombo.setEnabled(    state );
                                _tagField.setEnabled(      state );
                                _travelField.setEnabled(   state );
                                _speedField.setEnabled(    state );
                                _damageField.setEnabled(   state );
                                _iterField.setEnabled(     state );
                                _startCheck.setEnabled(    state );
                                
                                
                                for ( int i = 0; i < _labels.length; ++i ) {
                                        _labels[i].setEnabled( state );
                                }
                                

                        }

                        private Map.SectSpecial a_spec = null;
                        
                        private Map.ElementAction theAction = new Map.ElementAction() {

                                        public void doAction( Map.Element elem ) {
                                                if ( elem instanceof Map.Sector ) {

                                                        Map.Sector sect = (Map.Sector) elem;

                                                        if ( a_spec == null ) {
                                                                sect.deleteSpecial();
                                                                sect._map.addCacheClip( sect.getBounds( 0 ) );
                                                                return;
                                                        }
                                                        
                                                        if ( a_spec == sect._special )
                                                                return;

                                                        if ( sect._special == null )
                                                                sect.createNewSpecial();

                                                        sect._special._tag    = a_spec._tag;
                                                        sect._special._travel = a_spec._travel;
                                                        sect._special._speed  = a_spec._speed;
                                                        sect._special._damage = a_spec._damage;
                                                        sect._special._iterations = a_spec._iterations;
                                                        sect._special._startExtended = a_spec._startExtended;

                                                        sect._map.addCacheClip( sect.getBounds( 0 ) );
                                                }
                                        }
                                };

                        ActionListener _specListener = new ActionListener() {
                                        public void actionPerformed( ActionEvent evt ) {

                                                if ( _isSettingElement )
                                                        return;
 
                                                Object source = evt.getSource();

                                                Map.SectSpecial spec = _sect._special;
                                                        
                                                if ( source == _specialCheck ) {
                                                        if ( _specialCheck.isSelected() ) {
                                                                _sect.createNewSpecial();
                                                        }
                                                        else {
                                                                _sect.deleteSpecial();
                                                        }
                                                }
                                                
                                                else if ( spec != null ) {
                                                        
                                                        try {     
                                                                if ( source == _specialCombo ) {
                                                                        spec._type = (short) _specialCombo.getSelectedIndex();
                                                                }
                                                        
                                                                else if ( source == _soundCombo ) {
                                                                        spec._sound = (short)_soundCombo.getSelectedIndex();
                                                                }
                                                        
                                                                else if ( source == _tagField ) {
                                                                        spec._tag = Byte.parseByte( _tagField.getText() );
                                                                }
                                                                
                                                                else if ( source == _travelField ) {
                                                                        spec._travel = Short.parseShort( _travelField.getText() );
                                                                }
                                                                
                                                                else if ( source == _speedField ) {
                                                                        spec._speed = Byte.parseByte( _speedField.getText() );
                                                                }
                                                                
                                                                else if ( source == _damageField ) {
                                                                        spec._damage = Short.parseShort( _damageField.getText() );
                                                                }
                                                                else if ( source == _iterField ) {
                                                                        spec._iterations = Byte.parseByte( _iterField.getText() );
                                                                }
                                                                 else if ( source == _delayField ) {
                                                                        spec._delay = Byte.parseByte( _delayField.getText() );
                                                                }
                                                                else if ( source == _startCheck ) {
                                                                        spec._startExtended = _startCheck.isSelected();
                                                                }
                                                        }
                                                        catch ( NumberFormatException exp ) {
                                                                Toolkit.getDefaultToolkit().beep();
                                                                return;
                                                                // ignore
                                                        }
                                                }

                                                a_spec = _sect._special;
                                                _map.doSelectionActions( SpecialPanel.this.theAction );
                                                _map.addCacheClip( _sect.getBounds( 0 ) );

                                                _map.notifyListenersFromCachedClip();
                                        }
                                };
                }
        }

        public static class FocusHelper extends FocusAdapter {
                ActionListener _listener = null;

                public FocusHelper( ActionListener alist ) {
                        _listener = alist;
                }
                
                public void focusLost( FocusEvent evt ) {
                        if ( evt.isTemporary() )
                                return;
                        
                        _listener.actionPerformed( new ActionEvent( evt.getSource(),
                                                                    evt.getID(),
                                                                    "focus helper" ) );
                }
        };
}
