import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

public class MapUtil {

        public static class MakeStairsAction implements Map.ElementAction {
                private int _floorStart;
                private int _ceilingStart;
                
                private int _floorIncrement;
                private int _ceilingIncrement;
                
                private int _curFloor;
                private int _curCeiling;

                public MakeStairsAction( int floorStart, int floorIncrement,
                                         int ceilingStart, int ceilingIncrement ) {
                        
                        _floorStart       = floorStart;
                        _ceilingStart     = ceilingStart;
                        _curFloor         = floorStart;
                        _curCeiling       = ceilingStart;
                        _floorIncrement   = floorIncrement;
                        _ceilingIncrement = ceilingIncrement;
                }
                
                public void doAction( Map.Element elem ) {
                        if ( ! ( elem instanceof Map.Sector ) )
                                return;

                        Map.Sector s = (Map.Sector) elem;
                        
                        s.setFloorHeight( _curFloor, true );
                        s.setCeilingHeight( _curCeiling, true );

                        _curFloor   += _floorIncrement;
                        _curCeiling += _ceilingIncrement;
                }
        }

        public static void showStairsDialog( final Frame frame, final Map map ) {
                
                final JDialog dialog = new JDialog( frame, "Create Stairs", true );

                final JButton applyButton     = new JButton( "Apply" );
                final JButton okButton     = new JButton( "Ok" );
                final JButton cancelButton = new JButton( "Cancel" );
                
                final JTextField[] startFields = new JTextField[] {
                        new JTextField( 6 ),
                        new JTextField( 6 ) };
                
                final JTextField[] endFields = new JTextField[] {
                        new JTextField( 6 ),
                        new JTextField( 6 ) };
                
                final int[] startVals = new int[] { 0, 0 };
                final int[] endVals   = new int[] { 0, 0 };
                final int[] increments = new int[] { 0, 0 };
                
                final int numSelections = map.getNumSelections();
                
                Map.Element elem = null;
                
                elem = map.getFirstSelection();
                
                if ( elem instanceof Map.Sector ) {
                        Map.Sector s = (Map.Sector) elem;
                        startVals[0] = s.getFloorHeight();
                        startVals[1] = s.getCeilingHeight();
                }
                
                startFields[0].setText( "" + startVals[0] );
                startFields[1].setText( "" + startVals[1] );
                
                elem = map.getLastSelection();
                
                if ( elem instanceof Map.Sector ) {
                        Map.Sector s = (Map.Sector) elem;
                        endVals[0] = s.getFloorHeight();
                        endVals[1] = s.getCeilingHeight();
                }
                else {
                        endVals[0] = startVals[0];
                        endVals[1] = startVals[1];
                }
                
                endFields[0].setText(   "" + endVals[0] );
                endFields[1].setText( "" + endVals[1] );
                
                final int[] incrementSteps = new int[] {
                        4, 8, 16, 24, 32, 48, 64, 96, 128
                };
                
                Integer[] incrementIntegerSteps = new Integer[ incrementSteps.length * 2 + 1 ];
                
                int j = 0;
                for ( int i = incrementSteps.length-1; i >= 0; --i ) {
                        incrementIntegerSteps[ j++ ] = new Integer( - incrementSteps[i] );
                }
                
                incrementIntegerSteps[ j++ ] = new Integer( 0 );

                for ( int i = 0; i < incrementSteps.length; ++i ) {
                        incrementIntegerSteps[ j++ ] = new Integer( incrementSteps[i] );
                }
                
                final JComboBox[] combos = new JComboBox[] {
                        new JComboBox( incrementIntegerSteps ),
                        new JComboBox( incrementIntegerSteps )
                                };
                
                combos[0].setEditable( true );
                combos[1].setEditable( true );
                
                combos[0].getEditor().setItem( new Integer( endVals[0] - startVals[0] ) );
                combos[1].getEditor().setItem( new Integer( endVals[1] - startVals[1] ) );

                //
                // setup various listeners
                //
                
                ActionListener actionListener = new ActionListener() {
                                public void actionPerformed( ActionEvent evt ) {
                                        
                                        Object source = evt.getSource();
                                        
                                        try {
                                                for ( int i = 0; i < 2; ++i ) {
                                                        if ( source == combos[i] ) {
                                                                Object val = combos[i].getEditor().getItem();
                                                
                                                                if ( val instanceof String )
                                                                        increments[i] = Integer.parseInt( (String) val );
                                                                else if ( val instanceof Number )
                                                                        increments[i] = ( (Number) val ).intValue();
                                                                else {
                                                                        System.out.println( " bogus type " + val );
                                                                }
                                                
                                                                endVals[i] = startVals[i] + numSelections * increments[i];
                                                                endFields[i].setText(   "" + endVals[i] );
                                                                break;
                                                        }
                                                
                                                        if ( source == startFields[i] ) {
                                                        
                                                                startVals[i] = Integer.parseInt( startFields[i].getText() );
                                                                increments[i] = endVals[i] - startVals[i];
                                                                break;
                                                        }

                                                        if ( source == endFields[i] ) {
                                                                endVals[i] = Integer.parseInt( endFields[i].getText() );
                                                                increments[i] = endVals[i] - startVals[i];
                                                                break;
                                                        }
                                                
                                                }
                                        }
                                        catch ( NumberFormatException exp ) {
                                                Toolkit.getDefaultToolkit().beep();
                                        }
                                }
                        };
                
                for ( int i = 0; i < 2; ++i ) {
                        combos[i].addActionListener( actionListener );
                        endFields[i].addActionListener( actionListener );
                        startFields[i].addActionListener( actionListener );
                }
                
                Container c = dialog.getContentPane();
                
                c.setLayout( new GridBagLayout() );
                GridBagConstraints gbc = new GridBagConstraints();

                gbc.gridx = gbc.gridy = 0;
                gbc.fill = gbc.BOTH;
                gbc.weightx = gbc.weighty = 1;
                gbc.insets = new Insets( 4, 4, 4, 4 );
                
                ++gbc.gridx;
                c.add( new JLabel( "Start Level" ), gbc );

                ++gbc.gridx;
                c.add( new JLabel( "Increment" ), gbc );

                ++gbc.gridx;
                c.add( new JLabel( "End Level" ), gbc );
                
                gbc.gridx = 0;
                ++gbc.gridy;

                c.add( new JLabel( "Floor" ), gbc );
                
                ++gbc.gridx;
                c.add( startFields[0], gbc );

                ++gbc.gridx;
                c.add( combos[0], gbc );

                ++gbc.gridx;
                c.add( endFields[0], gbc );

                gbc.gridx = 0;
                ++gbc.gridy;

                c.add( new JLabel( "Ceiling" ), gbc );
                
                ++gbc.gridx;
                c.add( startFields[1], gbc );

                ++gbc.gridx;
                c.add( combos[1], gbc );

                ++gbc.gridx;
                c.add( endFields[1], gbc );

                gbc.gridx = 0;
                ++gbc.gridy;
                gbc.gridwidth = 4;
                gbc.anchor = gbc.EAST;
                gbc.weightx = 0;
                gbc.insets = new Insets( 8, 4, 4, 4 );

                JPanel buttonPanel = new JPanel( new FlowLayout( FlowLayout.RIGHT ) );
                
                ActionListener buttonListener = new ActionListener() {
                                public void actionPerformed( ActionEvent evt ) {
                                        Object source = evt.getSource();

                                        if ( source == applyButton || source == okButton ) {
                                                
                                                MakeStairsAction act = new MakeStairsAction( startVals[0],
                                                                                             increments[0],
                                                                                             startVals[1],
                                                                                             increments[1] );
                                                map.doSelectionActions( act );
                                                map.notifyListenersFromCachedClip();

                                                if ( source == applyButton )
                                                        return;
                                        }

                                        dialog.dispose();
                                }
                        };

                applyButton.addActionListener(  buttonListener );
                okButton.addActionListener(     buttonListener );
                cancelButton.addActionListener( buttonListener );

                buttonPanel.add( applyButton );
                buttonPanel.add( okButton );
                buttonPanel.add( cancelButton );

                c.add( buttonPanel, gbc );
                
                dialog.setLocationRelativeTo( frame );
                dialog.pack();
                dialog.show();
        }
}
