import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.awt.event.*;
import java.awt.*;
import java.util.*;

public class ScriptEditor extends JPanel implements Map.ScriptListener {
        
        ScriptTable _scriptTable;
        ScriptTableModel _scriptTableModel;
        
        CommandTable _cmdTable;
        CommandTableModel _cmdTableModel;

        MessageTable _msgTable;
        MessageTableModel _msgTableModel;

        Map _map;
        
        public ScriptEditor( Map map ) {
                _map = map;

                _scriptTableModel = new ScriptTableModel( _map );
                _cmdTableModel = new CommandTableModel();
                _msgTableModel = new MessageTableModel();
                
                _cmdTable = new CommandTable( _cmdTableModel );
                _scriptTable = new ScriptTable( _scriptTableModel );
                _msgTable = new MessageTable( _msgTableModel );

                _scriptTable.getSelectionModel().addListSelectionListener( new ListSelectionListener() {
                                public void valueChanged( ListSelectionEvent evt ) {
                                        int row = _scriptTable.getSelectedRow();

                                        Script s = row == -1 ? null : (Script) _map.scripts.elementAt( row );

                                        _cmdTableModel.setScript( s );
                                        _msgTableModel.setScript( s );
                                }
                        } );

                _map.scriptListeners.addElement( this );

                setLayout( new GridBagLayout() );

                GridBagConstraints gbc = new GridBagConstraints();

                gbc.insets = new Insets( 2, 2, 2, 2 );
                gbc.fill = gbc.BOTH;
                gbc.anchor = gbc.NORTHWEST;
                gbc.weightx = gbc.weighty = 1;
                gbc.gridx = gbc.gridy = 0;
                gbc.gridwidth = 2;

                gbc.weighty = 0;
                gbc.gridwidth = 1;
                add( new JLabel( "Scripts" ), gbc );
                
                ScriptButtons sb = new ScriptButtons( _scriptTable, _scriptTableModel );
                
                gbc.gridx = 1;
                add( sb, gbc );
                gbc.gridx = 0;
                gbc.gridwidth = 2;
                gbc.weighty = 1;
                
                gbc.weighty = 1;

                ++gbc.gridy;
                
                JScrollPane scroll = null;

                _scriptTable.setVisibleRowCount( 4 );
                scroll = new JScrollPane( _scriptTable );

                add( scroll, gbc );

                ++gbc.gridy;

                gbc.weighty = 0;
                gbc.gridwidth = 1;
                add( new JLabel( "Commands" ), gbc );
                
                CommandButtons cb = new CommandButtons( _cmdTable, _cmdTableModel );
                
                gbc.gridx = 1;
                add( cb, gbc );
                gbc.gridx = 0;
                gbc.gridwidth = 2;
                gbc.weighty = 1;
                
                ++gbc.gridy;


                _cmdTable.setVisibleRowCount( 15 );
                scroll = new JScrollPane( _cmdTable );
                add( scroll, gbc );
                                
                ++gbc.gridy;

                gbc.weighty = 0;
                gbc.gridwidth = 1;
                add( new JLabel( "Messages" ), gbc );

                MessageButtons mb = new MessageButtons( _msgTable, _msgTableModel );

                gbc.gridx = 1;
                add( mb, gbc );
                gbc.gridx = 0;
                gbc.gridwidth = 2;
                gbc.weighty = 1;

                ++gbc.gridy;


                _msgTable.setVisibleRowCount( 3 );
                scroll = new JScrollPane( _msgTable );
                add( scroll, gbc );


        }

        public void scriptsChanged() {
                _scriptTableModel.fireTableDataChanged();
        }
}


class CommandButtons extends JToolBar {

        CommandTable _table;
        CommandTableModel _model;
        
        CommandButtons( CommandTable table, CommandTableModel model ) {
                
                setFloatable( false );

                _table = table;
                _model = model;
                
                add( table.insertAction );
                add( table.appendAction );
                add( table.remAction );
                add( table.upAction );
                add( table.downAction );
        }
}

class ScriptButtons extends JToolBar {

        ScriptTable _table;
        ScriptTableModel _model;
        
        ScriptButtons( ScriptTable table, ScriptTableModel model ) {
                
                setFloatable( false );

                _table = table;
                _model = model;
                
                add( table.addAction );
                add( table.remAction );
        }
}

class MessageButtons extends JToolBar {

        MessageTable _table;
        MessageTableModel _model;
        
        MessageButtons( MessageTable table, MessageTableModel model ) {
                
                setFloatable( false );

                _table = table;
                _model = model;
                
                add( table.addAction );
                add( table.remAction );
        }
}

class ScriptTable extends DeluxeTable {
        
        AbstractAction addAction = new AbstractAction( "Add" ) {
                        public void actionPerformed( ActionEvent evt ) {
                                
                                int max_num = -1;

                                for ( Iterator iter = _model._map.scripts.iterator(); iter.hasNext(); ) {
                                        max_num = Math.max( max_num, ( (Script) iter.next() ).num );
                                }
                                
                                _model._map.scripts.addElement( new Script( max_num + 1 ) );
                                _model._map.notifyScriptListeners();
                        }
                };

        AbstractAction remAction = new AbstractAction( "Remove" ) {
                        public void actionPerformed( ActionEvent evt ) {
                                
                                if ( _model._map.scripts.size() == 0 ) {
                                        getToolkit().beep();
                                        return;
                                }

                                int row = getSelectedRow();
                                
                                if ( row >= 0 && _model._map.scripts.size() > row ) {
                                        Script script = (Script) _model._map.scripts.remove( row );
                                         
                                        boolean found = false;
                                        for ( Iterator iter = _model._map.lines.iterator(); iter.hasNext(); ) {
                                                Map.Line line = (Map.Line) iter.next();
                                                if ( line._script == script ) {
                                                        line._script = null;
                                                        _model._map.addCacheClip( line.getBounds( 0 ) );
                                                        found = true;
                                                }
                                        }
                                        
                                        if ( found )
                                                _model._map.notifyListenersFromCachedClip();

                                        if ( --row >= 0 )
                                                setRowSelectionInterval( row, row );

                                        _model._map.notifyScriptListeners();
                                }
                        }
                };


        ScriptTableModel _model;

        public ScriptTable( ScriptTableModel model ) {
                super( model );
                _model = model;
        }

}

class ScriptTableModel extends DeluxeTableModel {
        private static final String[] _colnames = new String[] {
                "Num",
                "Num Cmds",
                "Num Msgs",
        };

        private static final Class[] _classes = new Class[] {
                Integer.class,
                Integer.class,
                Integer.class,
        };

        Map _map;
        public ScriptTableModel( Map map ) {
                _map = map;
                System.out.println( "map has scriptps " + _map.scripts.size() );
        }

        public Class getColumnClass( int col ) { return _classes[ col ]; }
        public int getRowCount() { return _map.scripts.size(); }
        public int getColumnCount() { return 3; }
        public String getColumnName( int col ) { return _colnames[ col ]; }
        public Object getValueAt( int row, int col ) {
                Script script = (Script) _map.scripts.elementAt( row );

                switch ( col ) {
                case 0:
                        return new Integer( script.num );
                case 1:
                        return new Integer( script.cmds.size() );
                case 2:
                        return new Integer( script.msgs.size() );
                }

                return null;
        }

        public boolean isCellEditable( int row, int col ) {
                return col == 0;
        }

        public void setValueAt( Object value, int row, int col ) {
                try {
                        int i = Integer.parseInt( value.toString() );
                        Script script = (Script) _map.scripts.elementAt( row );
                        script.num = i;

                        boolean found = false;
                        for ( Iterator iter = _map.lines.iterator(); iter.hasNext(); ) {
                                Map.Line line = (Map.Line) iter.next();
                                if ( line._script == script ) {
                                        _map.addCacheClip( line.getBounds( 0 ) );
                                        found = true;
                                }
                        }
                                        
                        if ( found )
                                _map.notifyListenersFromCachedClip();

                }
                catch ( NumberFormatException exp ) {
                        Toolkit.getDefaultToolkit().beep();
                }
        }
}

class CommandTable extends DeluxeTable {

        AbstractAction insertAction = new AbstractAction( "Insert" ) {
                        public void actionPerformed( ActionEvent evt ) {

                                if ( _model._script == null ) {
                                        getToolkit().beep();
                                        return;
                                }
                                
                                int row = getSelectedRow();
                                
                                if ( row < 0 || _model._script.cmds.size() == 0 ) {
                                        row = 0;
                                }
                                
                                _model._script.cmds.add( row, new Script.Command() );
                                ++row;
                                setRowSelectionInterval( row, row );
                                _model.fireTableDataChanged();
                        }
                };

        AbstractAction appendAction = new AbstractAction( "Append" ) {
                        public void actionPerformed( ActionEvent evt ) {

                                if ( _model._script == null ) {
                                        getToolkit().beep();
                                        return;
                                }
                                int row = getSelectedRow();
                                
                                if ( row < 0 || _model._script.cmds.size() == 0 ) {
                                        row = _model._script.cmds.size() - 1;
                                }

                                ++row;
                                _model._script.cmds.add( row, new Script.Command() );
                                _model.fireTableDataChanged();
                              
                        }
                };

        AbstractAction remAction = new AbstractAction( "Remove" ) {
                        public void actionPerformed( ActionEvent evt ) {
                                
                                if ( _model._script == null || _model._script.cmds.size() == 0 ) {
                                        getToolkit().beep();
                                        return;
                                }

                                int row = getSelectedRow();
                                
                                if ( row >= 0 && _model._script.cmds.size() > row ) {
                                        _model._script.cmds.removeElementAt( row );
                                         
                                        if ( --row >= 0 )
                                                setRowSelectionInterval( row, row );

                                        _model.fireTableDataChanged(); 
                                }
                        }
                };
        AbstractAction upAction = new AbstractAction( "Up" ) {
                        public void actionPerformed( ActionEvent evt ) {
                                if ( _model._script == null ) {
                                        getToolkit().beep();
                                        return;
                                }
                                
                                int row = getSelectedRow();
                                
                                if ( row <= 0 || _model._script.cmds.size() < 2 ) {
                                        getToolkit().beep();
                                        return; 
                                }  

                                Script.Command cmd = (Script.Command) _model._script.cmds.remove( row );
                                --row;
                                _model._script.cmds.add( row, cmd );
                                setRowSelectionInterval( row, row );
                                _model.fireTableDataChanged(); 
                        }
                };
        AbstractAction downAction = new AbstractAction( "Down" ) {
                        public void actionPerformed( ActionEvent evt ) {
                                if ( _model._script == null ) {
                                        getToolkit().beep();
                                        return;
                                }
                                
                                int row = getSelectedRow();
                            
                                if ( row < 0 || row == ( _model._script.cmds.size() - 1 ) ) {
                                        getToolkit().beep();
                                        return; 
                                }  

                                Script.Command cmd = (Script.Command) _model._script.cmds.remove( row );
                                ++row;
                                _model._script.cmds.add( row, cmd );
                                setRowSelectionInterval( row, row );
                                _model.fireTableDataChanged(); 
                                
                        }
                };
        
        CommandTableModel _model = null;
        
        public CommandTable( CommandTableModel model ) {
                super( model );
                _model = model;
                TableColumn typecol = getColumnModel().getColumn(1);
                JComboBox typecombo = new JComboBox( Script.Command.typenames );
                typecol.setCellEditor( new DefaultCellEditor( typecombo ) );
                typecombo.setMaximumRowCount( 15 );

                autoSizeColumns();

        }
        
        public void doPopup( MouseEvent evt, int row, int col ) {
                System.out.println( "popup at " + row + "/ " + col );
        }
}

class CommandTableModel extends DeluxeTableModel {
        
        private static final Object[] _minWidths = new Object[] {
                "888",
                "XXXXXXXXXXXXXXX",
                null,
                null,
                null,
                null,
        };

        private static final Class[] _classes = new Class[] {
                Integer.class,
                String.class,
                Integer.class,
                Integer.class,
                Integer.class,
                Integer.class,
        };

        private static final String[] _colnames = new String[] {
                "Num",
                "Type",
                "Value 0",
                "Value 1",
                "Value 2",
                "Value 3",
        };

        Script _script = null;

        public CommandTableModel() {

        }

        void setScript( Script script ) {
                _script = script;
                fireTableDataChanged();
        }

        public Object getMinimumWidthColumnValue( int col ) { return _minWidths[ col ]; }
        public Class getColumnClass( int col ) { return _classes[ col ]; }
        public int getRowCount() { return _script == null ? 0 : _script.cmds.size(); }
        public int getColumnCount() { return _colnames.length; }
        public String getColumnName( int col ) { return _colnames[ col ]; }
        public boolean isCellEditable( int row, int col ) {
                if ( col > 0 )
                        return true;
                return false;
        }

        public Object getValueAt( int row, int col ) { 
                
                if ( _script == null )
                        return null;
                
                Script.Command cmd = (Script.Command) _script.cmds.elementAt( row );

                if ( cmd == null )
                        return null;
                
                switch ( col ) {
                case 0:
                        return new Integer( row );
                case 1:
                        return cmd.typenames[ cmd.type ];
                case 2:
                case 3:
                case 4:
                case 5:
                        return new Integer( cmd.values[ col - 2 ] );
                }

                return null;
        }
        
        public String getToolTipAt( int row, int col ) { 
                
                if ( _script == null )
                        return null;
                
                Script.Command cmd = (Script.Command) _script.cmds.elementAt( row );
                
                if ( cmd == null )
                        return null;
                
                switch ( col ) {
                case 2:
                case 3:
                case 4:
                case 5:
                        return cmd.valuenames[ cmd.type ][ col - 2 ];
                }

                return null;
        }

        public void setValueAt( Object value, int row, int col ) {
   
                if ( _script == null )
                        return;
                
                Script.Command cmd = (Script.Command) _script.cmds.elementAt( row );

                if ( cmd == null )
                        return;
                
                switch ( col ) {
                case 1:
                        for ( int i = 0; i < cmd.typenames.length; ++i )
                                if ( cmd.typenames[i].equals( value ) ) {
                                        cmd.type = (byte) i;
                                        break;
                                }
                        break;
                case 2:
                case 3:
                case 4:
                case 5:
                        try {
                                short s = Short.parseShort( value.toString() );
                                cmd.values[ col - 2 ] = s;
                        }
                        catch ( NumberFormatException exp ) {
                                Toolkit.getDefaultToolkit().beep();
                                System.out.println( exp );
                        }
                        break;
                }
        }

}

class MessageTable extends DeluxeTable {

        AbstractAction addAction = new AbstractAction( "Add" ) {
                        public void actionPerformed( ActionEvent evt ) {
                                
                                if ( _model._script == null ) {
                                        getToolkit().beep();
                                        return;
                                }
                                        
                                _model._script.msgs.addElement( "Default" );
                                _model.fireTableDataChanged();
                        }
                };

        AbstractAction remAction = new AbstractAction( "Remove" ) {
                        public void actionPerformed( ActionEvent evt ) {
                                
                                if ( _model._script == null ||  _model._script.msgs.size() == 0 ) {
                                        getToolkit().beep();
                                        return;
                                }

                                int row = getSelectedRow();
                                
                                if ( row >= 0 && _model._script.msgs.size() > row ) {
                                        _model._script.msgs.remove( row );
                                         
                                        if ( --row >= 0 )
                                                setRowSelectionInterval( row, row );
                                        _model.fireTableDataChanged();
                                }
                        }
                };


        MessageTableModel _model;

        public MessageTable( MessageTableModel model ) {
                super( model );
                _model = model;

                autoSizeColumns();
        }
}

class MessageTableModel extends DeluxeTableModel {
        
        private static final String[] _colnames = new String[] {
                "Num",
                "Text",
        };
        
        private static final Class[] _classes = new Class[] {
                Integer.class,
                String.class,
        };

        Script _script = null;

        public MessageTableModel( ) {
        }

        void setScript( Script script ) {
                _script = script;
                fireTableDataChanged();
        }

        public Class getColumnClass( int col ) { return _classes[ col ]; }
        public int getRowCount() { return _script == null ? 0 : _script.msgs.size(); }
        public int getColumnCount() { return _colnames.length; }
        public String getColumnName( int col ) { return _colnames[ col ]; }
        public Object getValueAt( int row, int col ) { 
                switch ( col ) {
                case 0:
                        return new Integer( row );
                case 1:
                        return _script.msgs.elementAt( row );
                }                        
                return null;
        }

        public boolean isColumnExpandable( int col ) {
                return col == 1;
        }

        public boolean isCellEditable( int row, int col ) {
                return col == 1;
        }

        public void setValueAt( Object value, int row, int col ) {
                if ( _script == null )
                        return;

                _script.msgs.set( row, value.toString() );
                //                fireTableDataChanged();
        }
}
       
