import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;

import java.awt.event.*;
import java.awt.*;
import java.util.*;

public class TextureChecker extends JPanel {

        private Map _map = null;
        private JCheckBox _sectorBox = new JCheckBox( "Check Sectors", true );
        private JCheckBox _sideBox  = new JCheckBox( "Check Sides", true );

        private Vector _probs = new Vector();
        private TexTableModel _model = new TexTableModel();
        private JTable _table = new JTable( _model );
        private JScrollPane _scroll = new JScrollPane( _table );

        private JButton _cancelButton = new JButton( "Cancel" );
        private JButton _okButton     = new JButton( "Ok" );
        private JButton _defaultButton = new JButton( "Apply Defaults" );

        ListSelectionModel _selModel = null;
        
        public TextureChecker( Map map ) {
                _map = map;
                
                JPanel panel = new JPanel();

                panel.add( _sectorBox );
                panel.add( _sideBox );
                panel.add( _cancelButton );
                panel.add( _okButton );
                panel.add( _defaultButton );
                
                setLayout( new BorderLayout() );
                
                add( panel, "North" );
                
                add( _scroll, "Center" );
                
                _table.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );

                setBorder( BorderFactory.createRaisedBevelBorder() );
                
                _table.getSelectionModel().addListSelectionListener( _selectionListener );
                _okButton.addActionListener( _checkListener );
                _defaultButton.addActionListener( _defaultListener );
        }
        
        ListSelectionListener _selectionListener = new ListSelectionListener() {
                        public void valueChanged( ListSelectionEvent evt ) {
                                if ( evt.getValueIsAdjusting() )
                                        return;

                                int index = _table.getSelectedRow();
                                
                                if ( index == -1 ) {
                                        _map.setSelection( null, false );
                                        return;
                                }

                                ProbInfo p = (ProbInfo ) _probs.elementAt( index );
                                
                                Object obj = null;
                                        
                                if ( p.isSect ) {
                                        obj = _map.sectors.elementAt( p.num );
                                }
                                else {
                                        _map.preferredLineSide = p.isFront ? 0 : 1;
                                        obj = _map.lines.elementAt( p.num );
                                }

                                _map.setSelection( (Map.Element) obj, false );
                                        
                                
                        }
                };
        
        ActionListener _defaultListener = new ActionListener() {
                        public void actionPerformed( ActionEvent evt ) {
                                int count = 0;
                                
                                for ( Iterator iter = _probs.iterator(); iter.hasNext(); ) {
                                        ProbInfo p = (ProbInfo) iter.next();

                                        if ( p.isSect ) {
                                                Map.Sector s = (Map.Sector) _map.sectors.elementAt( p.num );
                                                if ( p.which == s.FLOOR_TEX ) {
                                                        s.setTex( s.FLOOR_TEX, _map.getDefaultTex( _map.DEF_FLOOR_TEX ) );
                                                        ++count;
                                                }
                                                else {
                                                        s.setTex( s.CEILING_TEX, _map.getDefaultTex( _map.DEF_CEILING_TEX ) );
                                                        ++count;
                                                }
                                        }

                                        else {
                                                Map.Line line = (Map.Line) _map.lines.elementAt( p.num );
                                                Map.Side s = line.line_sides[ p.isFront ? 0 : 1 ];
                                                s.setTexture( p.which, _map.getDefaultTex( _map.DEF_WALL_TEX ) );
                                                ++count;
                                        }
                                }
                                
                                System.out.println( " :: Default texture applied at " + count + " locations." );
                        }
                };
        
        ActionListener _checkListener = new ActionListener() {
                        public void actionPerformed( ActionEvent evt ) {
                                _probs.clear();
                                
                                if ( _sectorBox.isSelected() ) {
                                        int index = 0;
                                
                                        for ( Iterator iter = _map.sectors.iterator(); iter.hasNext(); ++index ) {
                                                Map.Sector s = (Map.Sector) iter.next();
                                                
                                                if ( s.getTexIndex( s.FLOOR_TEX ) == 0 ) {
                                                        ProbInfo p = new ProbInfo( true, index, s.FLOOR_TEX, false );
                                                        _probs.addElement( p );
                                                }
                                                if ( s.getTexIndex( s.CEILING_TEX ) == 0 ) {
                                                        ProbInfo p = new ProbInfo( true, index, s.CEILING_TEX, false );
                                                        _probs.addElement( p );
                                                }
                                        }
                                }
                                
                                if ( _sideBox.isSelected() ) {
                                        int index = 0;
                                        for ( Iterator iter = _map.lines.iterator(); iter.hasNext(); ++index ) {
                                                
                                                Map.Line line = (Map.Line) iter.next();
                                                
                                                for ( int i = 0; i < 2; ++i ) {
                                                        
                                                        Map.Side side = line.line_sides[ i ];

                                                        if ( side == null )
                                                                continue;

                                                        Map.Side other = line.line_sides[ 1 - i ];
                                                        
                                                        // main
                                                        if ( side.hasSection( side.MAIN_INDEX, other ) ) {
                                                                if ( side.getTexture( side.MAIN_INDEX ) == 0 ) {
                                                                        ProbInfo p = new ProbInfo( false, index, side.MAIN_INDEX, i == 0 );
                                                                        _probs.addElement( p );
                                                                }
                                                                break;
                                                        }

                                                        // lower
                                                        if ( side.hasSection( side.LOWER_INDEX, other ) ) {
                                                                if ( side.getTexture( side.LOWER_INDEX ) == 0 ) {
                                                                        ProbInfo p = new ProbInfo( false, index, side.LOWER_INDEX, i == 0 );
                                                                        _probs.addElement( p );
                                                                }
                                                        }

                                                        // upper
                                                        if ( side.hasSection( side.UPPER_INDEX, other ) ) {
                                                                if ( side.getTexture( side.UPPER_INDEX ) == 0 ) {
                                                                        ProbInfo p = new ProbInfo( false, index, side.UPPER_INDEX, i == 0 );
                                                                        _probs.addElement( p );
                                                                }
                                                        }
                                                }
                                        }
                                }

                                System.out.println( " :: " + _probs.size() + " texture problems detected." );

                                _model.fireTableDataChanged();
                        }
                };

        class ProbInfo {
                String[] strs = new String[2];
                boolean isSect = false;
                int num = 0;
                int which = 0;
                boolean isFront = false;

                public ProbInfo( boolean is, int n, int w, boolean f ) {
                        isSect = is;
                        num = n;
                        which = w;
                        isFront = f;
                        if ( isSect ) {
                                strs[0] = "Sector " + num;
                                strs[1] = which == Map.Sector.FLOOR_TEX ? "Floor" : "Ceiling";
                        }
                        else {
                                strs[0] = "Line " + num;
                                strs[1] = ( isFront ? "Front " : "Back " ) + ( which == Map.Side.UPPER_INDEX ? "Upper" :
                                                                               ( which == Map.Side.MAIN_INDEX ? "Main" : "Lower" ) );
                        }
                }
        }
        
        private class TexTableModel extends AbstractTableModel {
                
                public TexTableModel() {}
                
                public int getRowCount() { return _probs.size(); }
                public int getColumnCount() { return 2; }

                final String[] _names = new String[] { "Type", "Location" };
                
                public String getColumnName( int col ) { return _names[col]; }

                public Object getValueAt( int row, int col ) {
                        ProbInfo p = (ProbInfo) _probs.elementAt( row );
                        
                        return p.strs[ col ];
                }
        }
}
