/*
 * View2dSide.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.awt.geom.*;

final public class View2dSide extends View2d {
        
        public int HZ_NONE = 0;
        public int HZ_X = 1;
        public int HZ_Y = 2;
        
        private int _type = HZ_X;
        
        public View2dSide( JMenu menu ) {
                super( menu );
                _showVertsItem.setEnabled( false );
                _showLineArrowsItem.setEnabled( false );
                _showSidesItem.setEnabled( false );
        }

        public boolean isType( int type ) {
                return _type == type;
        }

        public int getType() {
                return _type;
        }

        public void setType( int type ) {
                _type = type;
                repaint();
        }
        
        /**
         * center the view
         * @param centerOnSelection true if the selection(s) are used for centering
         * @param multiSelection true if the center point is the average of the selections
         *
         */
        public void centerView( boolean centerOnSelection, boolean multiSelection ) {

                if ( _scroll != null ) {
                        Dimension vs = _scroll.getViewport().getExtentSize();

                        Point tp = new Point( ( _panelSize.width - vs.width ) / 2,
                                              ( _panelSize.height - vs.height ) / 2 );

                        Point3D cp = new Point3D();
                                               
                        if ( centerOnSelection ) {
                                
                                //
                                // center on average of centerpoints
                                //

                                if ( multiSelection ) {
                                        cp = _map.findSelectionsCenter();
                                        
                                }

                                //
                                // center on the last selection only
                                //

                                else {
                                        Map.Element elem = _map.getLastSelection();
                                        if ( elem != null ) {
                                                cp = elem.getCenterPoint3D();
                                        }
                                }
                                
                                //
                                // y axis is horizontal
                                //
                                
                                if ( _type == HZ_Y )
                                        cp.x = cp.y;
                                
                                //
                                // z axis is always vertical
                                //

                                cp.y = cp.z;

                                cp.x *= _zoom;
                                cp.y *= _zoom;

                        }
                                                 
                        tp.x += cp.x;
                        tp.y -= cp.y;
                        
                        tp.x = Math.min( _panelSize.width - vs.width, tp.x );
                        tp.x = Math.max( 0, tp.x );
                        tp.y = Math.min( _panelSize.height - vs.height, tp.y );
                        tp.y = Math.max( 0, tp.y );

                        _scroll.getViewport().setViewPosition( tp );
                }
        }
        
        /**
         * the top view simply returns the map size as indicated by the map
         */
        public Dimension getMapSliceSize() {
                
                return new Dimension( _type == HZ_X ? _map.getMapSize().width : _map.getMapSize().height,
                                      _map.getMapHeightZ() );
                
        }


        void drawTransient( Graphics g ) {
                
                super.drawTransient( g );
                
                if ( transClipDrawn == false && _map._transClip != null ) {

                        if ( _type == HZ_X )
                                lastTransClip.setBounds( _map._transClip.x, - _map.getMapHeightZ() / 2,
                                                         _map._transClip.width, _map.getMapHeightZ() );
                        else
                                lastTransClip.setBounds( _map._transClip.y, - _map.getMapHeightZ() / 2, 
                                                         _map._transClip.height, _map.getMapHeightZ() );
                         
                        g.setXORMode( RUBBER_COLOR );
                        g.drawRect( lastTransClip.x,
                                    lastTransClip.y, 
                                    lastTransClip.width, 
                                    lastTransClip.height );
                        transClipDrawn = true;
                        g.setPaintMode();
                }

                if ( _map._rubberBand != null ) {
                        
                        Rectangle myRubber = 
                                new Rectangle( _type == HZ_X ? _map._rubberBand[0].x : _map._rubberBand[0].y, 
                                               _map._rubberBand[0].z,
                                               _type == HZ_X ? _map._rubberBand[1].x : _map._rubberBand[1].y,
                                               _map._rubberBand[1].z );
                        
                        lastRubberBand.setBounds( myRubber );
                        g.setXORMode( RUBBER_COLOR );
                        g.drawRect( lastRubberBand.x, 
                                    lastRubberBand.y, 
                                    lastRubberBand.width, 
                                    lastRubberBand.height );
                        rubberBandDrawn = true;
                        g.setPaintMode();
                }
        }
        

        private final Rectangle d_rect = new Rectangle();

        final void drawSector( Graphics g, 
                               Map.Sector s, 
                               Rectangle rect, 
                               boolean fill, 
                               boolean line ) {
                
                if ( fill ) {
                        g.fillRect( d_rect.x, d_rect.y, d_rect.width, d_rect.height );
                        
                        if ( line )
                                g.setColor( LINE_COLOR );
                }
                
                if ( line )
                        g.drawRect( d_rect.x, d_rect.y, d_rect.width, d_rect.height );

                if ( s._special != null ) {
                        final Color oldcolor = g.getColor();
                        d_rect.x += d_rect.width / 2;
                        int arrow_wid = 10;
                        final int base_wid = d_rect.width * 3 / 8;

                        if ( s._special._type == s._special.SECT_DOOR ) {
                                d_rect.height = s.ceilingHeight - s._special._travel;
                                d_rect.y      = s.ceilingHeight;
                        }
                        else {
                                d_rect.height = s.floorHeight + s._special._travel;
                                d_rect.y      = s.floorHeight;
                                arrow_wid = -arrow_wid;
                        }

                        g.setColor( Color.yellow ); 

                        g.drawLine( d_rect.x, d_rect.y, d_rect.x, d_rect.height );
                        g.drawLine( d_rect.x - base_wid, d_rect.height, d_rect.x + base_wid, d_rect.height );
                        g.drawLine( d_rect.x - arrow_wid, d_rect.height + arrow_wid, d_rect.x, d_rect.height );
                        g.drawLine( d_rect.x + arrow_wid, d_rect.height + arrow_wid, d_rect.x, d_rect.height );
                        
                        g.setColor( oldcolor );
                }
        }

        final void drawLight( final Graphics g, final int x, final int y ) {
                g.drawOval( x - 8, y - 8, 16, 16 );
                g.drawOval( x - 4, y - 4, 8, 8 );
        }
        
        //
        // ElementAction objects for drawing the selected items
        //
        
        abstract protected class DrawSelected implements Map.ElementAction {
                Graphics g;
                boolean fill = true;
                Rectangle clip = null;
                int extent_x;
                int extent_y;
        }
 
        DrawSelected _drawSelectedLightsAction = new DrawSelected() {
                        public void doAction( Map.Element elem ) {
                                if ( elem instanceof Map.LightSource ) {
                                        Map.LightSource light = (Map.LightSource) elem;

                                        if ( _type == HZ_X ) {
                                                if ( light.pos.x < clip.x || light.pos.x > extent_x ||
                                                     light.pos.z < clip.y || light.pos.z > extent_y )
                                                        return;
                                                drawLight( g, light.pos.x, light.pos.z );
                                        }
                                        
                                        else {
                                                if ( light.pos.y < clip.x || light.pos.y > extent_x ||
                                                     light.pos.z < clip.y || light.pos.z > extent_y )
                                                        return;
                                                drawLight( g, light.pos.y, light.pos.z );
                                                
                                        }
                                }
                        }
                };

        DrawSelected _drawSelectedSectsAction = new DrawSelected() {
                        public void doAction( Map.Element elem ) {
                                if ( elem instanceof Map.Sector ) {
                                        final Map.Sector s = (Map.Sector) elem;
                                        
                                        if ( s.ceilingHeight < clip.y ||
                                             s.floorHeight > extent_y ) 
                                                return;
                                        
                                        final Rectangle rect = s.getBounds( 0 );

                                        if ( _type == HZ_X ) {
                                                if ( rect.x > extent_x || ( rect.x + rect.width ) < clip.x )
                                                        return;
                                                
                                                d_rect.setBounds( rect.x,
                                                                  s.floorHeight,
                                                                  rect.width,
                                                                  s.ceilingHeight - s.floorHeight );
                                        }
                                        else {
                                                if ( rect.y > extent_x || ( rect.y + rect.height ) < clip.x )
                                                        return;
                                                
                                                d_rect.setBounds( rect.y,
                                                                  s.floorHeight,
                                                                  rect.height,
                                                                  s.ceilingHeight - s.floorHeight );
                                        }
                                        
                                        if ( fill ) {
                                                Color color =  
                                                        _fillSectAlphaItem.isSelected() ? 
                                                        ( s._special != null ? 
                                                          SELECTED_SPEC_SECT_FILL_ALPHA : 
                                                          SELECTED_SECT_FILL_ALPHA ) :
                                                        ( s._special != null ? 
                                                          SELECTED_SPEC_SECT_FILL : 
                                                          SELECTED_SECT_FILL );  
                                                g.setColor( color );
                                        }
                                        
                                        drawSector( g, s, rect, fill, true );
                                        
                                } 
                        }
                };

        /*
        final long[] timer_totals = new long[ 10 ];
        final long[] timers       = new long[ timer_totals.length + 1 ];
        int num_counts = 0;
        */

        public void paint( Graphics ga ) {
                
                final Graphics2D g = (Graphics2D) ga;
                
                g.setRenderingHint( RenderingHints.KEY_RENDERING,
                                    RenderingHints.VALUE_RENDER_SPEED );
                g.setRenderingHint( RenderingHints.KEY_ALPHA_INTERPOLATION,
                                    RenderingHints.VALUE_ALPHA_INTERPOLATION_SPEED );
                g.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
                                    RenderingHints.VALUE_ANTIALIAS_OFF);
                g.setRenderingHint( RenderingHints.KEY_COLOR_RENDERING, 
                                    RenderingHints.VALUE_COLOR_RENDER_SPEED );
                g.setStroke( nilStroke );

                super.paint( g );

                /*
                int timer_i = 0;
                timers[ timer_i++ ] = System.currentTimeMillis();
                */

                final Dimension slice = getMapSliceSize();
                final Rectangle sliceRect = new Rectangle( - slice.width / 2,
                                                         - slice.height / 2,
                                                         slice.width,
                                                         slice.height );
                g.scale( _zoom, _zoom );
                g.transform( _displayTransform );

                final Rectangle user_clip = new Rectangle( _type == HZ_X ? _map._clip.x : _map._clip.y,
                                                           sliceRect.y,
                                                           _type == HZ_X ? _map._clip.width : _map._clip.height,
                                                           sliceRect.height );
                                                           
                final Rectangle big_clip = g.getClipBounds().intersection( sliceRect );
                
                final Rectangle clip = big_clip.intersection( user_clip );
                  
                //
                // draw the underlay grid
                //

                g.setClip( clip );
                
                if ( _showGridItem.isSelected() && _gridUnderItem.isSelected() ) {
                        drawGrid( g );
                }
                
              
                final int extent_x = clip.x + clip.width;
                final int extent_y = clip.y + clip.height;

                final int depth = 
                        _type == HZ_X ?
                        ( _map._clip.y + _map._clip.height ) :
                        ( _map._clip.x + _map._clip.width );
                
                //
                // draw the underlay grid
                //

                if ( _showGridItem.isSelected() && _gridUnderItem.isSelected() ) {
                        drawGrid( g );
                }
                
                final Map.Element selectedElem = _map.getLastSelection();
                final Class selectedClass = ( selectedElem != null ? selectedElem.getClass() : null );
                
                //
                // draw all unselected sectors
                //

                if ( _fillSectItem.isSelected() || _showLinesItem.isSelected() ) {
                        Color defColor =  
                                _fillSectAlphaItem.isSelected() ? 
                                SECT_FILL_ALPHA :
                                SECT_FILL;
                        
                        if ( _fillSectItem.isSelected() == false )
                                defColor = LINE_COLOR;
                        
                        g.setColor( defColor );

                        for ( Iterator iter = _map.sectors.iterator(); iter.hasNext(); ) {
                                Map.Sector s = (Map.Sector) iter.next();
                                
                                if ( ! s.isSelected ) {

                                        if ( s.ceilingHeight < clip.y ||
                                             s.floorHeight > extent_y ) 
                                                continue;
                                        
                                        final Rectangle rect = s.getBounds( 0 );

                                        if ( _type == HZ_X ) {

                                                if ( rect.x > extent_x || ( rect.x + rect.width ) < clip.x ||
                                                     rect.y > depth || ( rect.y + rect.height ) < _map._clip.y )
                                                        continue;

                                                d_rect.setBounds( rect.x,
                                                                  s.floorHeight,
                                                                  rect.width, 
                                                                  s.ceilingHeight - s.floorHeight );
                                        }
                                        else {
                                                
                                                if ( rect.y > extent_x || ( rect.y + rect.height ) < clip.x ||
                                                     rect.x > depth || ( rect.x + rect.width ) < _map._clip.x )

                                                        continue;

                                                d_rect.setBounds( rect.y, 
                                                                  s.floorHeight,
                                                                  rect.height, 
                                                                  s.ceilingHeight - s.floorHeight );
                                        }
                                        
                                        if ( _fillSectItem.isSelected() ) {
                                                if ( s._special != null ) {
                                                      g.setColor( _fillSectAlphaItem.isSelected() ? 
                                                                  SPEC_SECT_FILL_ALPHA :
                                                                  SPEC_SECT_FILL );    
                                                }
                                                else {
                                                        g.setColor( defColor );
                                                }
                                        }
                                        
                                        drawSector( g, s, rect,
                                                    _fillSectItem.isSelected(),
                                                    _showLinesItem.isSelected() );
                                }
                        }
                }
                 
                //
                // draw the selected sector(s) if needed
                //
                
                if ( Map.Sector.class == selectedClass ) {
                        _drawSelectedSectsAction.g = g;
                        _drawSelectedSectsAction.fill = true;
                        _drawSelectedSectsAction.clip = clip;
                        _drawSelectedSectsAction.extent_x = extent_x;
                        _drawSelectedSectsAction.extent_y = extent_y;
                        
                        if ( _drawSelectedSectsAction.fill == false ) {
                                g.setColor( LINE_COLOR );
                        }

                        _map.doSelectionActions( _drawSelectedSectsAction );
                }
                
                //
                // draw all unselected light sources
                //

                if ( true ) {
                        
                        g.setColor( LIGHT_SOURCE_COLOR );
                        
                        for ( Iterator iter = _map.lights.iterator(); iter.hasNext(); ) {
                                Map.LightSource light = (Map.LightSource) iter.next();
                                
                                if ( light.isSelected )
                                        continue;
                                
                                if ( _type == HZ_X ) {
                                        if ( light.pos.x < clip.x || light.pos.x > extent_x ||
                                             light.pos.z < clip.y || light.pos.z > extent_y ||
                                             light.pos.y > depth || light.pos.y < _map._clip.y )
                                                continue;
                                        drawLight( g, light.pos.x, light.pos.z );
                                }
                                
                                else {
                                        if ( light.pos.y < clip.x || light.pos.y > extent_x ||
                                             light.pos.z < clip.y || light.pos.z > extent_y ||
                                             light.pos.x > depth || light.pos.x < _map._clip.x )
                                                continue;
                                        drawLight( g, light.pos.y, light.pos.z );
                                }
                        }
                }
                
                //
                // draw the selected sector(s) if needed
                //
                
                if ( Map.LightSource.class == selectedClass ) {
                        _drawSelectedLightsAction.g = g;
                        _drawSelectedLightsAction.clip = clip;
                        _drawSelectedLightsAction.extent_x = extent_x;
                        _drawSelectedLightsAction.extent_y = extent_y;

                        g.setColor( SELECTED_COLOR );
                        _map.doSelectionActions( _drawSelectedLightsAction );
                }

                //
                // draw the overlay grid if applicable
                //

                if ( _showGridItem.isSelected() && _gridOverItem.isSelected() ) {
                        drawGrid( g );
                }
                
                transDrawn = false;
                rubberBandDrawn = false;
                transClipDrawn = false;

                g.setClip( big_clip );

                //
                // draw the user defined clip region on top of everything
                //

                g.setColor( Color.yellow );
                
                final int ce_x = user_clip.x + user_clip.width;
                final int ce_y = user_clip.y + user_clip.height; 

                g.drawLine( user_clip.x, big_clip.y,
                            user_clip.x, big_clip.y + big_clip.height );
                g.drawLine( ce_x, big_clip.y,
                            ce_x, big_clip.y + big_clip.height );
                g.drawLine( big_clip.x, user_clip.y,
                            big_clip.x + big_clip.width, user_clip.y );
                g.drawLine( big_clip.x, ce_y,
                            big_clip.x + big_clip.width, ce_y );
                
                // transient

                drawTransient( g );

                /*
                timers[ timer_i++ ] = System.currentTimeMillis();
                
                for ( int i = 1; i < timer_i; ++i ) {
                        timer_totals[i-1] = timers[i] - timers[ i-1 ];
                }
                ++num_counts;
                
                if ( num_counts > 30 ) {
                        System.out.println( " ------- results " );
                        for ( int i = 0; i < timer_i-1; ++i ) {
                                //                                timer_totals[i] /= num_counts;
                                System.out.println( "       segment " + i + ": " + timer_totals[i] );
                                timer_totals[i] = 0;
                        }
                        num_counts = 0;
                }
                */
        }
        public void mapChanged( Rectangle clip ) {

                //                System.out.println( " ++ MAP CHANGED w clip=" + clip );

                if ( clip != null ) {
                        Dimension d = getMapSliceSize();
                        Rectangle rect;
                        
                        if ( _type == HZ_X ) {
                                
                                rect = new Rectangle( (int) ( ( clip.x + d.width / 2 ) * _zoom ),
                                                      0,
                                                      (int) ( clip.width * _zoom ),
                                                      _panelSize.height );
                        }
                        else {
                                rect = new Rectangle( (int) ( ( clip.y + d.width / 2 ) * _zoom ),
                                                      0,
                                                      (int) ( clip.height * _zoom ),
                                                      _panelSize.height );
                        }
                        
                        rect.grow( 2, 2 );
                        
                        repaint( rect );
                }
                else
                        repaint();
        }
        
}
