/*
 * View2dTop.java:
 *
 * Copyright (C) 2000 John Watson, jwatson@tempusmud.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.awt.geom.*;

public final class View2dTop extends View2d {
        
        public View2dTop( JMenu menu ) {
                super( menu );
        }
        
        /**
         * center the view
         * @param centerOnSelection true if the selection(s) are used for centering
         * @param multiSelection true if the center point is the average of the selections
         *
         */
        public void centerView( boolean centerOnSelection, boolean multiSelection ) {
                if ( _scroll != null ) {
                        Dimension vs = _scroll.getViewport().getExtentSize();
                        
                        Point tp = new Point( ( _panelSize.width - vs.width ) / 2,
                                              ( _panelSize.height - vs.height ) / 2 );

                        Point cp = new Point();
                                               
                        if ( centerOnSelection ) {
                                
                                //
                                // center on average of centerpoints
                                //

                                if ( multiSelection ) {
                                        cp = _map.findSelectionsCenter();
                                }

                                //
                                // center on the last selection only
                                //

                                else {
                                        Map.Element elem = _map.getLastSelection();
                                        if ( elem != null ) {
                                                cp = elem.getCenterPoint();
                                        }
                                }

                                cp.x *= _zoom;
                                cp.y *= _zoom;

                        }
                                                 
                        tp.x += cp.x;
                        tp.y -= cp.y;
                        
                        tp.x = Math.min( _panelSize.width - vs.width, tp.x );
                        tp.x = Math.max( 0, tp.x );
                        tp.y = Math.min( _panelSize.height - vs.height, tp.y );
                        tp.y = Math.max( 0, tp.y );

                        _scroll.getViewport().setViewPosition( tp );
                        

                }
        }
        
        /**
         * the top view simply returns the map size as indicated by the map
         */
        public Dimension getMapSliceSize() {
                
                return _map.getMapSize();
        }

        void drawTransient( Graphics g ) {
                
                super.drawTransient( g );

                if ( transClipDrawn == false && _map._transClip != null ) {
                        lastTransClip.setBounds( _map._transClip );
                        
                        g.setXORMode( RUBBER_COLOR );
                        g.drawRect( _map._transClip.x,
                                    _map._transClip.y, 
                                    _map._transClip.width, 
                                    _map._transClip.height );
                        transClipDrawn = true;
                        g.setPaintMode();
                }

                if ( _map._rubberBand != null ) {
                        
                        Rectangle myRubber = new Rectangle( _map._rubberBand[0].x, _map._rubberBand[0].y,
                                                            _map._rubberBand[1].x, _map._rubberBand[1].y );
                        
                        lastRubberBand.setBounds( myRubber );
                        g.setXORMode( RUBBER_COLOR );
                        g.drawRect( lastRubberBand.x, 
                                    lastRubberBand.y, 
                                    lastRubberBand.width, 
                                    lastRubberBand.height );
                        rubberBandDrawn = true;
                        g.setPaintMode();
                }

                if ( transDrawn ) {
                        g.setXORMode( Color.white );
                        g.drawLine( lastTrans[0].x, lastTrans[0].y,
                                    lastTrans[1].x, lastTrans[1].y );
                        transDrawn = false;
                        g.setPaintMode();
                }
                
                if ( _map.drawTransient ) {
                        g.setXORMode( Color.white );
                        g.drawLine( _map.tranPts[0].x, _map.tranPts[0].y, _map.tranPts[1].x, _map.tranPts[1].y );
                        g.setPaintMode();
                        transDrawn = true;
                        lastTrans[0].setLocation( _map.tranPts[0] );
                        lastTrans[1].setLocation( _map.tranPts[1] );
                }

        }

        final private void drawLight( final Graphics g, final Map.LightSource light ) {
                g.drawOval( light.pos.x - 8, light.pos.y - 8, 16, 16 );
                g.drawOval( light.pos.x - 4, light.pos.y - 4, 8, 8 );
        }
        
        private double[] un = new double[2];
        private double[] uv = new double[2];
        private int[] arrow = new int[4];
        private int[] mid  = new int[2];

        void drawMapLine( Graphics g, Map.Line line, Color color ) {
                          
                g.setColor( color );
                g.drawLine( line.verts[0].x, line.verts[0].y,
                            line.verts[1].x, line.verts[1].y );

                //
                // unit vector and normal for line
                //

                VMath.unitNormal( line.verts[1].x - line.verts[0].x,
                                  line.verts[1].y - line.verts[0].y, un );
                
                //
                // draw the arrow point
                //

                if ( _showLineArrowsItem.isSelected() ) {

                        uv[0] = - un[1];
                        uv[1] =   un[0];
                        
                        arrow[0] = (int) ( line.verts[1].x - uv[0] * 12 + un[0] *  8 );
                        arrow[1] = (int) ( line.verts[1].y - uv[1] * 12 + un[1] *  8 );
                        arrow[2] = (int) ( line.verts[1].x - uv[0] * 12 - un[0] *  8 );
                        arrow[3] = (int) ( line.verts[1].y - uv[1] * 12 - un[1] *  8 );
                        
                        g.drawLine( line.verts[1].x, line.verts[1].y,
                                    arrow[0], arrow[1] );

                        g.drawLine( line.verts[1].x, line.verts[1].y,
                                    arrow[2], arrow[3] );
                }
                
                //
                // draw the sides
                //
                
                if ( _showSidesItem.isSelected() ) {
                        if ( line.line_sides[0] != null ) {
                                double[][] side_pts = new double[][] {
                                        { line.verts[0].x + un[0] * 3,
                                          line.verts[0].y + un[1] * 3 },
                                        { line.verts[1].x + un[0] * 3,
                                          line.verts[1].y + un[1] * 3 }
                                };
                 
                                g.setColor( SIDE_COLOR );
                                g.drawLine( (int) side_pts[0][0], (int) side_pts[0][1],
                                            (int) side_pts[1][0], (int) side_pts[1][1] );
                        }
                
                        if ( line.line_sides[1] != null ) {
                        
                                double[][] side_pts = new double[][] {
                                        { line.verts[0].x - un[0] * 3,
                                          line.verts[0].y - un[1] * 3 },
                                        { line.verts[1].x - un[0] * 3,
                                          line.verts[1].y - un[1] * 3 }
                                };

                                g.setColor( SIDE_COLOR );
                                g.drawLine( (int) side_pts[0][0], (int) side_pts[0][1],
                                            (int) side_pts[1][0], (int) side_pts[1][1] );
                        }

                        g.setColor( color );
                }
                
                //
                // draw the little normal
                //

                mid[0] = ( line.verts[1].x + line.verts[0].x ) / 2;
                mid[1] = ( line.verts[1].y + line.verts[0].y ) / 2;

                g.drawLine( mid[0], mid[1], (int) ( mid[0] + un[0] * 10 ), (int) ( mid[1] + un[1] * 10 ) );

        }
        
        //
        // ElementAction objects for drawing the selected items
        //
        
        abstract protected class DrawSelected implements Map.ElementAction {
                Graphics g;
        }
 
        DrawSelected _drawSelectedSectsAction = new DrawSelected() {
                        public void doAction( Map.Element elem ) {
                                if ( elem instanceof Map.Sector ) {
                                        Map.Sector sect = (Map.Sector) elem;
                                        
                                        Color color =  
                                                _fillSectAlphaItem.isSelected() ? 
                                                ( sect._special != null ? 
                                                  SELECTED_SPEC_SECT_FILL_ALPHA : 
                                                  SELECTED_SECT_FILL_ALPHA ) :
                                                ( sect._special != null ? 
                                                  SELECTED_SPEC_SECT_FILL : 
                                                  SELECTED_SECT_FILL );
                                        
                                        g.setColor( color );
                                        g.fillPolygon( sect.getPolygon() );
                                }
                        }
                };

        DrawSelected _drawSelectedLinesAction = new DrawSelected() {
                        public void doAction( Map.Element elem ) {
                                if ( elem instanceof Map.Line ) {
                                        Map.Line line = (Map.Line) elem;
                                        drawMapLine( g, line, 
                                                     line._script == null ? 
                                                     SELECTED_COLOR :
                                                     SPEC_LINE_SELECTED_COLOR );
                                }
                        }
                };

        DrawSelected _drawSelectedVertsAction = new DrawSelected() {
                        public void doAction( Map.Element elem ) {
                                if ( elem instanceof Map.Vertex ) {
                                        Map.Vertex v = (Map.Vertex) elem;
                                        g.fillRect( v.x - 4, v.y - 4, 8, 8 );
                                }
                        }
                };

        DrawSelected _drawSelectedLightsAction = new DrawSelected() {
                        public void doAction( Map.Element elem ) {
                                if ( elem instanceof Map.LightSource ) {
                                        drawLight( g, (Map.LightSource) elem );
                                }
                        }
                };

        /*
        final long[] timer_totals = new long[ 10 ];
        final long[] timers       = new long[ timer_totals.length + 1 ];
        int num_counts = 0;
        */
        public void paint( Graphics ga ) {
                
                final Graphics2D g = (Graphics2D) ga;
                
                g.setRenderingHint( RenderingHints.KEY_RENDERING,
                                    RenderingHints.VALUE_RENDER_SPEED );
                g.setRenderingHint( RenderingHints.KEY_ALPHA_INTERPOLATION,
                                    RenderingHints.VALUE_ALPHA_INTERPOLATION_SPEED );
                g.setRenderingHint( RenderingHints.KEY_ANTIALIASING,
                                    RenderingHints.VALUE_ANTIALIAS_OFF);
                g.setRenderingHint( RenderingHints.KEY_COLOR_RENDERING, 
                                    RenderingHints.VALUE_COLOR_RENDER_SPEED );

                g.setStroke( nilStroke );

                
                // int timer_i = 0;
                //                timers[ timer_i++ ] = System.currentTimeMillis();

                super.paint( g );

                final Rectangle mapRect = _map.getMapRect();
                
                g.scale( _zoom, _zoom );
                g.transform( _displayTransform );
                
                final Rectangle big_clip = g.getClipBounds().intersection( mapRect );
                final Rectangle clip = big_clip.intersection( _map._clip );

                if ( big_clip.width == 0 || big_clip.height == 0 ) {
                        System.out.println( " 2d top clip is ZERO" );
                        return;
                }

                //
                // draw the underlay grid
                //

                g.setClip( clip );
                
                if ( _showGridItem.isSelected() && _gridUnderItem.isSelected() ) {
                        drawGrid( g );
                }
                
                //
                // commence with drawing the map elements
                //

                final int extent_x = clip.x + clip.width;
                final int extent_y = clip.y + clip.height;
               
             
                final Map.Element selectedElem = _map.getLastSelection();
                final Class selectedClass = ( selectedElem != null ? selectedElem.getClass() : null );
                
                //
                // draw all unselected sectors
                //

                if ( _fillSectItem.isSelected() ) {
                        
                        final Color defColor =  
                                _fillSectAlphaItem.isSelected() ? 
                                SECT_FILL_ALPHA :
                                SECT_FILL;
                
                        g.setColor( defColor );
                
                        for ( Iterator iter = _map.sectors.iterator(); iter.hasNext(); ) {
                                final Map.Sector s = (Map.Sector) iter.next();

                                if ( ! s.isSelected ) {

                                        if ( s._special != null ) {
                                                g.setColor( _fillSectAlphaItem.isSelected() ? 
                                                            SPEC_SECT_FILL_ALPHA :
                                                            SPEC_SECT_FILL );
                                                g.fillPolygon( s.getPolygon() );
                                                g.setColor( defColor );
                                        }
                                        else
                                                g.fillPolygon( s.getPolygon() );
                                }
                        }
                }
                
                //
                // draw the selected sector(s) if needed
                //
                
                if ( Map.Sector.class == selectedClass ) {
                        _drawSelectedSectsAction.g = g;
                        _map.doSelectionActions( _drawSelectedSectsAction );
                }
                
                //
                // draw all unselected lines
                //

                if ( _showLinesItem.isSelected() ) {

                        for ( Iterator iter = _map.lines.iterator(); iter.hasNext(); ) {
                                final Map.Line line = (Map.Line) iter.next();
                                
                                
                                if ( ( line.verts[0].x < clip.x &&
                                       line.verts[1].x < clip.x ) ||
                                     ( line.verts[0].y < clip.y &&
                                       line.verts[1].y < clip.y ) ||
                                     ( line.verts[0].x > extent_x &&
                                       line.verts[1].x > extent_x ) ||
                                     ( line.verts[0].y > extent_y &&
                                       line.verts[1].y > extent_y ) )
                                        continue;

                                if ( ! line.isSelected ) {
                                        drawMapLine( g, line,
                                                     line._script == null ? 
                                                     LINE_COLOR :
                                                     SPEC_LINE_COLOR );
                                }
                        }
                }

                //
                // draw the selected line if needed
                //
                
                if ( Map.Line.class == selectedClass ) {
                        _drawSelectedLinesAction.g = g;
                        _map.doSelectionActions( _drawSelectedLinesAction );
                }

                //timers[ timer_i++ ] = System.currentTimeMillis();

                //
                // draw the vertexes
                //
                
                if ( _showVertsItem.isSelected() ) {
                        g.setColor( VERTS_COLOR );
                        
                        for ( Iterator iter = _map.vertexes.iterator(); iter.hasNext(); ) {
                                final Map.Vertex vert = (Map.Vertex) iter.next();
                                
                                if ( vert.x < clip.x ||
                                     vert.y < clip.y ||
                                     vert.x > extent_x ||
                                     vert.y > extent_y )
                                        continue;
                                
                                g.fillRect( vert.x - 4, vert.y - 4, 8, 8 );
                        }
                }
                
                //
                // draw the selected vertexes if needed
                //

                if ( Map.Vertex.class == selectedClass ) {
                        _drawSelectedVertsAction.g = g;
                        g.setColor( SELECTED_COLOR );
                        _map.doSelectionActions( _drawSelectedVertsAction );
                }

                //
                // draw all unselected light sources
                //

                if ( true ) {
                        
                        g.setColor( LIGHT_SOURCE_COLOR );
                        
                        for ( Iterator iter = _map.lights.iterator(); iter.hasNext(); ) {
                                final Map.LightSource light = (Map.LightSource) iter.next();

                                if ( light.pos.x < clip.x ||
                                     light.pos.y < clip.y ||
                                     light.pos.x > extent_x ||
                                     light.pos.y > extent_y )
                                        continue;
                                
                                if ( ! light.isSelected ) {
                                        drawLight( g, light );
                                }
                        }
                }
                
                //
                // draw the selected light(s) if needed
                //
                
                if ( Map.LightSource.class == selectedClass ) {
                        _drawSelectedLightsAction.g = g;
                        g.setColor( SELECTED_COLOR );
                        _map.doSelectionActions( _drawSelectedLightsAction );
                }

                //
                // draw the overlay grid if applicable
                //

                if ( _showGridItem.isSelected() && _gridOverItem.isSelected() ) {
                        drawGrid( g );
                }

                //
                // draw the in-progress new zone
                //


                if ( _map.numPtsInProg > 0 ) {
                        g.setColor( NEW_LINE_COLOR );
                        
                        g.drawPolyline( _map.ptsInProg[0],
                                        _map.ptsInProg[1],
                                        _map.numPtsInProg );

                        //
                        // draw the default closing line as it exists right now
                        //

                        if ( _map.numPtsInProg >= 3 ) {
                                g.setColor( NEW_SECT_DEFAULT_LINE_COLOR );
                                
                                g.drawLine( _map.ptsInProg[0][0], _map.ptsInProg[1][0],
                                            _map.ptsInProg[0][ _map.numPtsInProg - 1 ], 
                                            _map.ptsInProg[1][ _map.numPtsInProg - 1 ] );
                                g.setColor( NEW_LINE_COLOR );
                        }

                        for ( int i = 0; i < _map.numPtsInProg; ++i ) {
                                g.fillRect( _map.ptsInProg[0][i] - 4,
                                            _map.ptsInProg[1][i] - 4,
                                            8, 8 );
                        }
                }

                transDrawn = false;
                rubberBandDrawn = false;
                transClipDrawn = false;

                g.setClip( big_clip );

                //
                // draw the user defined clip region on top of everything
                //

                g.setColor( Color.yellow );
                
                final int ce_x = _map._clip.x + _map._clip.width;
                final int ce_y = _map._clip.y + _map._clip.height; 

                g.drawLine( _map._clip.x, big_clip.y,
                            _map._clip.x, big_clip.y + big_clip.height );
                g.drawLine( ce_x, big_clip.y,
                            ce_x, big_clip.y + big_clip.height );
                g.drawLine( big_clip.x, _map._clip.y,
                            big_clip.x + big_clip.width, _map._clip.y );
                g.drawLine( big_clip.x, ce_y,

                            big_clip.x + big_clip.width, ce_y );
                
                // transients

                drawTransient( g );

                
                /*
                timers[ timer_i++ ] = System.currentTimeMillis();

                for ( int i = 1; i < timer_i; ++i ) {
                        timer_totals[i-1] = timers[i] - timers[ i-1 ];
                }
                ++num_counts;

                if ( num_counts > 30 ) {
                        System.out.println( " ------- results " );
                        for ( int i = 0; i < timer_i-1; ++i ) {
                                //                                timer_totals[i] /= num_counts;
                                System.out.println( "       segment " + i + ": " + timer_totals[i] );
                                timer_totals[i] = 0;
                        }
                        num_counts = 0;
                }
                */
        }

        public void mapChanged( Rectangle clip ) {

                //                System.out.println( "MAPCHANGED with clip=" + clip + ", zoom=" + _zoom );

                if ( clip != null ) {
                        Dimension d = getMapSliceSize();

                        Rectangle rect = new Rectangle( (int) ( ( clip.x + d.width / 2 ) * _zoom ),
                                                        (int) ( ( - clip.y - clip.height + d.height / 2 ) * _zoom ) ,
                                                        (int) ( clip.width * _zoom ),
                                                        (int) ( clip.height *  _zoom ) );
                        rect.grow( 2, 2 );
                        
                        repaint( rect );
                }
                else
                        repaint();
        }

}
